from __future__ import division
import math
import warnings

from albumentations.core.utils import DataProcessor

__all__ = ['check_keypoints', 'convert_keypoints_from_albumentations', 'convert_keypoints_to_albumentations',
           'filter_keypoints', 'KeypointsProcessor']

keypoint_formats = {'xy', 'yx', 'xya', 'xys', 'xyas', 'xysa'}


class KeypointsProcessor(DataProcessor):
    @property
    def default_data_name(self):
        return "keypoints"

    def ensure_data_valid(self, data):
        if self.params.label_fields:
            if not all(l in data.keys() for l in self.params.label_fields):
                raise ValueError("Your 'label_fields' are not valid - them must have same names as params in "
                                 "'keypoint_params' dict")

    def ensure_transforms_valid(self, transforms):
        # IAA-based augmentations supports only transformation of xy keypoints.
        # If your keypoints formats is other than 'xy' we emit warning to let user
        # be aware that angle and size will not be modified.

        from albumentations.imgaug.transforms import DualIAATransform
        if self.params.format is not None and self.params.format != 'xy':
            for transform in transforms:
                if isinstance(transform, DualIAATransform):
                    warnings.warn("{} transformation supports only 'xy' keypoints "
                                  "augmentation. You have '{}' keypoints format. Scale "
                                  "and angle WILL NOT BE transformed.".format(transform.__class__.__name__,
                                                                              self.params.format))
                    break

    def filter(self, data, rows, cols):
        return filter_keypoints(data, rows, cols, remove_invisible=self.params.remove_invisible)

    def check(self, data, rows, cols):
        return check_keypoints(data, rows, cols)

    def convert_from_albumentations(self, data, rows, cols):
        return convert_keypoints_from_albumentations(data, self.params.format,
                                                     rows, cols,
                                                     check_validity=self.params.remove_invisible,
                                                     angle_in_degrees=self.params.angle_in_degrees)

    def convert_to_albumentations(self, data, rows, cols):
        return convert_keypoints_to_albumentations(data, self.params.format,
                                                   rows, cols,
                                                   check_validity=self.params.remove_invisible,
                                                   angle_in_degrees=self.params.angle_in_degrees)


def check_keypoint(kp, rows, cols):
    """Check if keypoint coordinates are in range [0, 1)"""
    for name, value, size in zip(['x', 'y'], kp[:2], [cols, rows]):
        if not 0 <= value < size:
            raise ValueError(
                'Expected {name} for keypoint {kp} '
                'to be in the range [0.0, {size}], got {value}.'.format(
                    kp=kp,
                    name=name,
                    value=value,
                    size=size
                )
            )


def check_keypoints(keypoints, rows, cols):
    """Check if keypoints boundaries are in range [0, 1)"""
    for kp in keypoints:
        check_keypoint(kp, rows, cols)


def filter_keypoints(keypoints, rows, cols, remove_invisible):
    if not remove_invisible:
        return keypoints

    resulting_keypoints = []
    for kp in keypoints:
        x, y = kp[:2]
        if x < 0 or x >= cols:
            continue
        if y < 0 or y >= rows:
            continue
        resulting_keypoints.append(kp)
    return resulting_keypoints


def keypoint_has_extra_data(kp, format):
    return len(kp) > len(format)


def convert_keypoint_to_albumentations(keypoint, source_format, rows, cols,
                                       check_validity=False, angle_in_degrees=True):
    if source_format not in keypoint_formats:
        raise ValueError(
            "Unknown target_format {}. Supported formats are: {}".format(source_format, keypoint_formats)
        )

    if source_format == 'xy':
        x, y = keypoint[:2]
        a, s = 0, 0
        tail = list(keypoint[2:])
    elif source_format == 'yx':
        y, x = keypoint[:2]
        a, s = 0, 0
        tail = list(keypoint[2:])
    elif source_format == 'xya':
        x, y, a = keypoint[:3]
        s = 0
        tail = list(keypoint[3:])
    elif source_format == 'xys':
        x, y, s = keypoint[:3]
        a = 0
        tail = list(keypoint[3:])
    elif source_format == 'xyas':
        x, y, a, s = keypoint[:4]
        tail = list(keypoint[4:])
    elif source_format == 'xysa':
        x, y, s, a = keypoint[:4]
        tail = list(keypoint[4:])

    if angle_in_degrees:
        a = math.radians(a)

    keypoint = [x, y, a, s] + tail
    if check_validity:
        check_keypoint(keypoint, rows, cols)
    return keypoint


def normalize_angle(a):
    two_pi = 2. * math.pi
    while a < 0:
        a += two_pi
    while a > two_pi:
        a -= two_pi
    return a


def convert_keypoint_from_albumentations(keypoint, target_format, rows, cols,
                                         check_validity=False, angle_in_degrees=True):
    if target_format not in keypoint_formats:
        raise ValueError(
            "Unknown target_format {}. Supported formats are: {}".format(target_format, keypoint_formats)
        )
    if check_validity:
        check_keypoint(keypoint, rows, cols)
    x, y, a, s = keypoint[:4]
    a = normalize_angle(a)
    if angle_in_degrees:
        a = math.degrees(a)

    if target_format == 'xy':
        kp = [x, y]
    elif target_format == 'yx':
        kp = [y, x]
    elif target_format == 'xya':
        kp = [x, y, a]
    elif target_format == 'xys':
        kp = [x, y, s]
    elif target_format == 'xyas':
        kp = [x, y, a, s]
    elif target_format == 'xysa':
        kp = [x, y, s, a]

    return kp + list(keypoint[4:])


def convert_keypoints_to_albumentations(keypoints, source_format, rows, cols,
                                        check_validity=False, angle_in_degrees=True):
    return [convert_keypoint_to_albumentations(kp, source_format, rows, cols, check_validity,
                                               angle_in_degrees=angle_in_degrees) for kp in keypoints]


def convert_keypoints_from_albumentations(keypoints, target_format, rows, cols,
                                          check_validity=False, angle_in_degrees=True):
    return [convert_keypoint_from_albumentations(kp, target_format, rows, cols, check_validity,
                                                 angle_in_degrees=angle_in_degrees) for kp in keypoints]



from __future__ import division, print_function
import argparse
import math
import os
import sys
from abc import ABC
from timeit import Timer
from collections import defaultdict
import pkg_resources

from Augmentor import Operations, Pipeline
from PIL import Image
import cv2
from tqdm import tqdm
import numpy as np
import pandas as pd
import torchvision.transforms.functional as torchvision
import keras as _
import keras_preprocessing.image as keras
from imgaug import augmenters as iaa
import solt.core as slc
import solt.transforms as slt
import solt.data as sld

import albumentations.augmentations.functional as albumentations


def parse_args():
    parser = argparse.ArgumentParser(description='Augmentation libraries performance benchmark')
    parser.add_argument('-d', '--data-dir', required=True, metavar='DIR', help='path to a directory with images')
    parser.add_argument('-i', '--images', default=2000, type=int, metavar='N',
                        help='number of images for benchmarking (default: 2000)')
    parser.add_argument('-r', '--runs', default=5, type=int, metavar='N',
                        help='number of runs for each benchmark (default: 5)')
    parser.add_argument('--show-std', dest='show_std', action='store_true',
                        help='show standard deviation for benchmark runs')
    parser.add_argument('-p', '--print-package-versions', action='store_true', help='print versions of packages')
    return parser.parse_args()


def print_package_versions():
    packages = [
        'albumentations',
        'imgaug',
        'torchvision',
        'keras',
        'numpy',
        'opencv-python',
        'scikit-image',
        'scipy',
        'pillow',
        'pillow-simd',
        'Augmentor',
        'solt',
    ]
    package_versions = {'python': sys.version}
    for package in packages:
        try:
            package_versions[package] = pkg_resources.get_distribution(package).version
        except pkg_resources.DistributionNotFound:
            pass
    print(package_versions)


def read_img_pillow(path):
    with open(path, 'rb') as f:
        img = Image.open(f)
        return img.convert('RGB')


def read_img_cv2(filepath):
    img = cv2.imread(filepath)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return img


def format_results(images_per_second_for_aug, show_std=False):
    if images_per_second_for_aug is None:
        return '-'
    result = str(math.floor(np.mean(images_per_second_for_aug)))
    if show_std:
        result += ' ± {}'.format(math.ceil(np.std(images_per_second_for_aug)))
    return result


class BenchmarkTest(ABC):

    def __str__(self):
        return self.__class__.__name__

    def imgaug(self, img):
        return self.imgaug_transform.augment_image(img)

    def augmentor(self, img):
        return self.augmentor_op.perform_operation([img])

    def solt(self, img):
        dc = sld.DataContainer(img, 'I')
        dc = self.solt_stream(dc)
        return dc.data[0]

    def is_supported_by(self, library):
        if library == 'imgaug':
            return hasattr(self, 'imgaug_transform')
        elif library == 'augmentor':
            return hasattr(self, 'augmentor_op') or hasattr(self, 'augmentor_pipeline')
        elif library == 'solt':
            return hasattr(self, 'solt_stream')
        else:
            return hasattr(self, library)

    def run(self, library, imgs):
        transform = getattr(self, library)
        for img in imgs:
            transform(img)


class HorizontalFlip(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.Fliplr(p=1)
        self.augmentor_op = Operations.Flip(probability=1, top_bottom_left_right='LEFT_RIGHT')
        self.solt_stream = slc.Stream([slt.RandomFlip(p=1, axis=1)])

    def albumentations(self, img):
        if img.ndim == 3 and img.shape[2] > 1 and img.dtype == np.uint8:
            return albumentations.hflip_cv2(img)
        else:
            return albumentations.hflip(img)

    def torchvision(self, img):
        return torchvision.hflip(img)

    def keras(self, img):
        return np.ascontiguousarray(keras.flip_axis(img, axis=1))

    def imgaug(self, img):
        return np.ascontiguousarray(self.imgaug_transform.augment_image(img))


class VerticalFlip(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.Flipud(p=1)
        self.augmentor_op = Operations.Flip(probability=1, top_bottom_left_right='TOP_BOTTOM')
        self.solt_stream = slc.Stream([slt.RandomFlip(p=1, axis=0)])

    def albumentations(self, img):
        return albumentations.vflip(img)

    def torchvision(self, img):
        return torchvision.vflip(img)

    def keras(self, img):
        return np.ascontiguousarray(keras.flip_axis(img, axis=0))

    def imgaug(self, img):
        return np.ascontiguousarray(self.imgaug_transform.augment_image(img))


class Rotate(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.Affine(rotate=(45, 45), order=1, mode='reflect')
        self.augmentor_op = Operations.RotateStandard(probability=1, max_left_rotation=45, max_right_rotation=45)
        self.solt_stream = slc.Stream([slt.RandomRotate(p=1, rotation_range=(45, 45))], padding='r')

    def albumentations(self, img):
        return albumentations.rotate(img, angle=-45)

    def torchvision(self, img):
        return torchvision.rotate(img, angle=-45, resample=Image.BILINEAR)

    def keras(self, img):
        return keras.apply_affine_transform(img, theta=45, channel_axis=2, fill_mode='reflect')


class Brightness(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.Add((127, 127), per_channel=False)
        self.augmentor_op = Operations.RandomBrightness(probability=1, min_factor=1.5, max_factor=1.5)
        self.solt_stream = slc.Stream([slt.ImageRandomBrightness(p=1, brightness_range=(127, 127))])

    def albumentations(self, img):
        return albumentations.brightness_contrast_adjust(img, beta=0.5)

    def torchvision(self, img):
        return torchvision.adjust_brightness(img, brightness_factor=1.5)

    def keras(self, img):
        return keras.apply_brightness_shift(img, brightness=1.5).astype(np.uint8)


class Contrast(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.Multiply((1.5, 1.5), per_channel=False)
        self.augmentor_op = Operations.RandomContrast(probability=1, min_factor=1.5, max_factor=1.5)
        self.solt_stream = slc.Stream([slt.ImageRandomContrast(p=1, contrast_range=(1.5, 1.5))])

    def albumentations(self, img):
        return albumentations.brightness_contrast_adjust(img, alpha=1.5)

    def torchvision(self, img):
        return torchvision.adjust_contrast(img, contrast_factor=1.5)


class BrightnessContrast(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.Sequential([
            iaa.Multiply((1.5, 1.5), per_channel=False),
            iaa.Add((127, 127), per_channel=False),
        ])
        self.augmentor_pipeline = Pipeline()
        self.augmentor_pipeline.add_operation(
            Operations.RandomBrightness(probability=1, min_factor=1.5, max_factor=1.5),
        )
        self.augmentor_pipeline.add_operation(
            Operations.RandomContrast(probability=1, min_factor=1.5, max_factor=1.5)
        )
        self.solt_stream = slc.Stream([
            slt.ImageRandomBrightness(p=1, brightness_range=(127, 127)),
            slt.ImageRandomContrast(p=1, contrast_range=(1.5, 1.5)),
        ])

    def albumentations(self, img):
        return albumentations.brightness_contrast_adjust(img, alpha=1.5, beta=0.5)

    def torchvision(self, img):
        img = torchvision.adjust_brightness(img, brightness_factor=1.5)
        img = torchvision.adjust_contrast(img, contrast_factor=1.5)
        return img

    def augmentor(self, img):
        for operation in self.augmentor_pipeline.operations:
            img, = operation.perform_operation([img])
        return img


class ShiftScaleRotate(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.Affine(
            scale=(2, 2),
            rotate=(45, 45),
            translate_px=(50, 50),
            order=1,
            mode='reflect',
        )

    def albumentations(self, img):
        return albumentations.shift_scale_rotate(img, angle=-45, scale=2, dx=0.2, dy=0.2)

    def torchvision(self, img):
        return torchvision.affine(img, angle=45, translate=(50, 50), scale=2, shear=0, resample=Image.BILINEAR)

    def keras(self, img):
        return keras.apply_affine_transform(img, theta=45, tx=50, ty=50, zx=0.5, zy=0.5, fill_mode='reflect')


class ShiftHSV(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.AddToHueAndSaturation((20, 20), per_channel=False)
        self.solt_stream = slc.Stream([slt.ImageRandomHSV(p=1, h_range=(20, 20), s_range=(20, 20), v_range=(20, 20))])

    def albumentations(self, img):
        return albumentations.shift_hsv(img, hue_shift=20, sat_shift=20, val_shift=20)

    def torchvision(self, img):
        img = torchvision.adjust_hue(img, hue_factor=0.1)
        img = torchvision.adjust_saturation(img, saturation_factor=1.2)
        img = torchvision.adjust_brightness(img, brightness_factor=1.2)
        return img


class Solarize(BenchmarkTest):

    def __init__(self):
        pass

    def albumentations(self, img):
        return albumentations.solarize(img)


class RandomCrop64(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.CropToFixedSize(width=64, height=64)
        self.augmentor_op = Operations.Crop(probability=1, width=64, height=64, centre=False)
        self.solt_stream = slc.Stream([slt.CropTransform(crop_size=(64, 64), crop_mode='r')])

    def albumentations(self, img):
        return albumentations.random_crop(img, crop_height=64, crop_width=64, h_start=0, w_start=0)

    def torchvision(self, img):
        return torchvision.crop(img, i=0, j=0, h=64, w=64)


class ShiftRGB(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.Add((100, 100), per_channel=False)

    def albumentations(self, img):
        return albumentations.shift_rgb(img, r_shift=100, g_shift=100, b_shift=100)

    def keras(self, img):
        return keras.apply_channel_shift(img, intensity=100, channel_axis=2)


class PadToSize512(BenchmarkTest):

    def __init__(self):
        self.solt_stream = slc.Stream([slt.PadTransform(pad_to=(512, 512), padding='r')])

    def albumentations(self, img):
        return albumentations.pad(img, min_height=512, min_width=512)

    def torchvision(self, img):
        if img.size[0] < 512:
            img = torchvision.pad(img, (int((1 + 512 - img.size[0]) / 2), 0), padding_mode='reflect')
        if img.size[1] < 512:
            img = torchvision.pad(img, (0, int((1 + 512 - img.size[1]) / 2)), padding_mode='reflect')
        return img


class Resize512(BenchmarkTest):

    def __init__(self):
        self.imgaug_transform = iaa.Scale(size=512, interpolation='linear')
        self.solt_stream = slc.Stream([slt.ResizeTransform(resize_to=(512, 512))])
        self.augmentor_op = Operations.Resize(probability=1, width=512, height=512, resample_filter='BILINEAR')

    def albumentations(self, img):
        return albumentations.resize(img, height=512, width=512)

    def torchvision(self, img):
        return torchvision.resize(img, (512, 512))


class Gamma(BenchmarkTest):

    def __init__(self):
        self.solt_stream = slc.Stream([slt.ImageGammaCorrection(p=1, gamma_range=(0.5, 0.5))])

    def albumentations(self, img):
        return albumentations.gamma_transform(img, gamma=0.5)

    def torchvision(self, img):
        return torchvision.adjust_gamma(img, gamma=0.5)


class Grayscale(BenchmarkTest):

    def __init__(self):
        self.augmentor_op = Operations.Greyscale(probability=1)
        self.imgaug_transform = iaa.Grayscale(alpha=1.0)
        self.solt_stream = slc.Stream([slt.ImageColorTransform(mode='rgb2gs')])

    def albumentations(self, img):
        return albumentations.to_gray(img)

    def torchvision(self, img):
        return torchvision.to_grayscale(img, num_output_channels=3)


def main():
    args = parse_args()
    if args.print_package_versions:
        print_package_versions()
    images_per_second = defaultdict(dict)
    libraries = [
        'albumentations',
        'imgaug',
        'torchvision',
        'keras',
        'augmentor',
        'solt',
    ]
    data_dir = args.data_dir
    paths = list(sorted(os.listdir(data_dir)))
    paths = paths[:args.images]
    imgs_cv2 = [read_img_cv2(os.path.join(data_dir, path)) for path in paths]
    imgs_pillow = [read_img_pillow(os.path.join(data_dir, path)) for path in paths]
    for library in libraries:
        imgs = imgs_pillow if library in ('torchvision', 'augmentor') else imgs_cv2
        benchmarks = [
            HorizontalFlip(),
            VerticalFlip(),
            Rotate(),
            ShiftScaleRotate(),
            Brightness(),
            Contrast(),
            BrightnessContrast(),
            ShiftRGB(),
            ShiftHSV(),
            Gamma(),
            Grayscale(),
            RandomCrop64(),
            PadToSize512(),
            Resize512(),
        ]
        pbar = tqdm(total=len(benchmarks))
        for benchmark in benchmarks:
            pbar.set_description('Current benchmark: {} | {}'.format(library, benchmark))
            benchmark_images_per_second = None
            if benchmark.is_supported_by(library):
                timer = Timer(lambda: benchmark.run(library, imgs))
                run_times = timer.repeat(number=1, repeat=args.runs)
                benchmark_images_per_second = [1 / (run_time / args.images) for run_time in run_times]
            images_per_second[library][str(benchmark)] = benchmark_images_per_second
            pbar.update(1)
        pbar.close()
    pd.set_option('display.width', 1000)
    df = pd.DataFrame.from_dict(images_per_second)
    df = df.applymap(lambda r: format_results(r, args.show_std))
    df = df[libraries]
    augmentations = ['RandomCrop64', 'PadToSize512', 'Resize512', 'HorizontalFlip', 'VerticalFlip', 'Rotate',
                     'ShiftScaleRotate', 'Brightness', 'Contrast', 'BrightnessContrast', 'ShiftHSV', 'ShiftRGB',
                     'Gamma', 'Grayscale']
    df = df.reindex(augmentations)
    print(df.head(len(augmentations)))


if __name__ == '__main__':
    main()



from albumentations import HorizontalFlip, OneOf, IAAPiecewiseAffine, IAAAffine, OpticalDistortion, GridDistortion
from albumentations.core.composition import Transforms
import pytest


def empty_aug1():
    return [
        HorizontalFlip(p=0.001),
        # IAAPiecewiseAffine(p=1.0),
        OneOf([
            # OpticalDistortion(p=0.1),
            # GridDistortion(p=0.1),
            # IAAPerspective(p=1.0),
            # IAAAffine(p=1.0),
            IAAPiecewiseAffine(p=1.0),
        ], p=0.0)
    ]


def empty_aug2():
    return [
        HorizontalFlip(p=0.001),
        IAAPiecewiseAffine(p=1.0),
        OneOf([
            # OpticalDistortion(p=0.1),
            # GridDistortion(p=0.1),
            # IAAPerspective(p=1.0),
            IAAAffine(p=1.0),
            # IAAPiecewiseAffine(p=1.0),
        ], p=0.0)
    ]


def empty_aug3():
    return [
        # HorizontalFlip(p=0.001),
        # IAAPiecewiseAffine(p=1.0),
        OneOf([
            OpticalDistortion(p=0.1),
            GridDistortion(p=0.1),
            # IAAPerspective(p=1.0),
            # IAAAffine(p=1.0),
            # IAAPiecewiseAffine(p=1.0),
        ], p=0.0)
    ]


@pytest.mark.parametrize(['aug', 'start_end'], [
    [empty_aug1, [0, 1]],
    [empty_aug2, [0, 2]],
    [empty_aug3, [0, 0]],
])
def test_strong_aug(aug, start_end):
    assert Transforms(aug()).start_end == start_end



import random

import cv2
import numpy as np
import pytest

from albumentations import (
    RandomCrop, PadIfNeeded, VerticalFlip, HorizontalFlip, Flip, Transpose,
    RandomRotate90, Rotate, ShiftScaleRotate, CenterCrop, OpticalDistortion,
    GridDistortion, ElasticTransform, RandomGridShuffle, ToGray, RandomGamma, JpegCompression,
    HueSaturationValue, RGBShift, Blur, MotionBlur, MedianBlur, GaussianBlur,
    GaussNoise, CLAHE, ChannelShuffle, InvertImg, IAAEmboss, IAASuperpixels,
    IAASharpen, IAAAdditiveGaussianNoise, IAAPiecewiseAffine, IAAPerspective,
    Cutout, CoarseDropout, Normalize, ToFloat, FromFloat,
    RandomBrightnessContrast, RandomSnow, RandomRain, RandomFog,
    RandomSunFlare, RandomCropNearBBox, RandomShadow, RandomSizedCrop,
    ChannelDropout, ISONoise, Solarize)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [JpegCompression, {}],
    [HueSaturationValue, {}],
    [RGBShift, {}],
    [RandomBrightnessContrast, {}],
    [Blur, {}],
    [MotionBlur, {}],
    [MedianBlur, {}],
    [GaussianBlur, {}],
    [GaussNoise, {}],
    [CLAHE, {}],
    [ChannelShuffle, {}],
    [InvertImg, {}],
    [RandomGamma, {}],
    [ToGray, {}],
    [Cutout, {}],
    [CoarseDropout, {}],
    [GaussNoise, {}],
    [RandomSnow, {}],
    [RandomRain, {}],
    [RandomFog, {}],
    [RandomSunFlare, {}],
    [RandomShadow, {}],
    [ChannelDropout, {}],
    [ISONoise, {}],
    [Solarize, {}],
])
def test_image_only_augmentations(augmentation_cls, params, image, mask):
    aug = augmentation_cls(p=1, **params)
    data = aug(image=image, mask=mask)
    assert data['image'].dtype == np.uint8
    assert data['mask'].dtype == np.uint8
    assert np.array_equal(data['mask'], mask)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [HueSaturationValue, {}],
    [RGBShift, {}],
    [RandomBrightnessContrast, {}],
    [Blur, {}],
    [MotionBlur, {}],
    [MedianBlur, {'blur_limit': (3, 5)}],
    [GaussianBlur, {}],
    [GaussNoise, {}],
    [ChannelShuffle, {}],
    [InvertImg, {}],
    [RandomGamma, {}],
    [JpegCompression, {}],
    [ToGray, {}],
    [Cutout, {}],
    [CoarseDropout, {}],
    [GaussNoise, {}],
    [RandomSnow, {}],
    [RandomRain, {}],
    [RandomFog, {}],
    [RandomSunFlare, {}],
    [RandomShadow, {}],
    [ChannelDropout, {}],
    [Solarize, {}]
])
def test_image_only_augmentations_with_float_values(augmentation_cls, params, float_image, mask):
    aug = augmentation_cls(p=1, **params)
    data = aug(image=float_image, mask=mask)
    assert data['image'].dtype == np.float32
    assert data['mask'].dtype == np.uint8
    assert np.array_equal(data['mask'], mask)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [PadIfNeeded, {}],
    [VerticalFlip, {}],
    [HorizontalFlip, {}],
    [Flip, {}],
    [Transpose, {}],
    [RandomRotate90, {}],
    [Rotate, {}],
    [ShiftScaleRotate, {}],
    [OpticalDistortion, {}],
    [GridDistortion, {}],
    [ElasticTransform, {}],
    [CenterCrop, {'height': 10, 'width': 10}],
    [RandomCrop, {'height': 10, 'width': 10}],
    [RandomSizedCrop, {'min_max_height': (4, 8), 'height': 10, 'width': 10}],
    [ISONoise, {}],
    [RandomGridShuffle, {}]
])
def test_dual_augmentations(augmentation_cls, params, image, mask):
    aug = augmentation_cls(p=1, **params)
    data = aug(image=image, mask=mask)
    assert data['image'].dtype == np.uint8
    assert data['mask'].dtype == np.uint8


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [PadIfNeeded, {}],
    [VerticalFlip, {}],
    [HorizontalFlip, {}],
    [Flip, {}],
    [Transpose, {}],
    [RandomRotate90, {}],
    [Rotate, {}],
    [ShiftScaleRotate, {}],
    [OpticalDistortion, {}],
    [GridDistortion, {}],
    [ElasticTransform, {}],
    [CenterCrop, {'height': 10, 'width': 10}],
    [RandomCrop, {'height': 10, 'width': 10}],
    [RandomSizedCrop, {'min_max_height': (4, 8), 'height': 10, 'width': 10}],
    [RandomGridShuffle, {}]
])
def test_dual_augmentations_with_float_values(augmentation_cls, params, float_image, mask):
    aug = augmentation_cls(p=1, **params)
    data = aug(image=float_image, mask=mask)
    assert data['image'].dtype == np.float32
    assert data['mask'].dtype == np.uint8


@pytest.mark.parametrize('augmentation_cls', [IAAEmboss, IAASuperpixels, IAASharpen, IAAAdditiveGaussianNoise])
def test_imgaug_image_only_augmentations(augmentation_cls, image, mask):
    aug = augmentation_cls(p=1)
    data = aug(image=image, mask=mask)
    assert data['image'].dtype == np.uint8
    assert data['mask'].dtype == np.uint8
    assert np.array_equal(data['mask'], mask)


@pytest.mark.parametrize('augmentation_cls', [IAAPiecewiseAffine, IAAPerspective])
def test_imgaug_dual_augmentations(augmentation_cls, image, mask):
    aug = augmentation_cls(p=1)
    data = aug(image=image, mask=mask)
    assert data['image'].dtype == np.uint8
    assert data['mask'].dtype == np.uint8


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [Cutout, {}],
    [JpegCompression, {}],
    [HueSaturationValue, {}],
    [RGBShift, {}],
    [RandomBrightnessContrast, {}],
    [RandomBrightnessContrast, {}],
    [Blur, {}],
    [MotionBlur, {}],
    [MedianBlur, {}],
    [GaussianBlur, {}],
    [GaussNoise, {}],
    [CLAHE, {}],
    [ChannelShuffle, {}],
    [InvertImg, {}],
    [RandomGamma, {}],
    [ToGray, {}],
    [Cutout, {}],
    [CoarseDropout, {}],
    [PadIfNeeded, {}],
    [VerticalFlip, {}],
    [HorizontalFlip, {}],
    [Flip, {}],
    [Transpose, {}],
    [RandomRotate90, {}],
    [Rotate, {}],
    [ShiftScaleRotate, {}],
    [OpticalDistortion, {}],
    [GridDistortion, {}],
    [ElasticTransform, {}],
    [CenterCrop, {'height': 10, 'width': 10}],
    [RandomCrop, {'height': 10, 'width': 10}],
    [RandomSizedCrop, {'min_max_height': (4, 8), 'height': 10, 'width': 10}],
    [Normalize, {}],
    [GaussNoise, {}],
    [ToFloat, {}],
    [FromFloat, {}],
    [RandomSnow, {}],
    [RandomRain, {}],
    [RandomFog, {}],
    [RandomSunFlare, {}],
    [RandomShadow, {}],
    [ChannelDropout, {}],
    [ISONoise, {}],
    [RandomGridShuffle, {}],
    [Solarize, {}],
])
def test_augmentations_wont_change_input(augmentation_cls, params, image, mask):
    image_copy = image.copy()
    mask_copy = mask.copy()
    aug = augmentation_cls(p=1, **params)
    aug(image=image, mask=mask)
    assert np.array_equal(image, image_copy)
    assert np.array_equal(mask, mask_copy)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [Cutout, {}],
    [CoarseDropout, {}],
    [HueSaturationValue, {}],
    [RGBShift, {}],
    [RandomBrightnessContrast, {}],
    [RandomBrightnessContrast, {}],
    [Blur, {}],
    [MotionBlur, {}],
    [MedianBlur, {'blur_limit': (3, 5)}],
    [GaussianBlur, {}],
    [GaussNoise, {}],
    [ChannelShuffle, {}],
    [InvertImg, {}],
    [RandomGamma, {}],
    [ToGray, {}],
    [PadIfNeeded, {}],
    [VerticalFlip, {}],
    [HorizontalFlip, {}],
    [Flip, {}],
    [Transpose, {}],
    [RandomRotate90, {}],
    [Rotate, {}],
    [ShiftScaleRotate, {}],
    [OpticalDistortion, {}],
    [GridDistortion, {}],
    [ElasticTransform, {}],
    [CenterCrop, {'height': 10, 'width': 10}],
    [RandomCrop, {'height': 10, 'width': 10}],
    [RandomSizedCrop, {'min_max_height': (4, 8), 'height': 10, 'width': 10}],
    [Normalize, {}],
    [GaussNoise, {}],
    [ToFloat, {}],
    [FromFloat, {}],
    [RandomSnow, {}],
    [RandomRain, {}],
    [RandomFog, {}],
    [RandomSunFlare, {}],
    [RandomShadow, {}],
    [ChannelDropout, {}],
    [RandomGridShuffle, {}],
    [Solarize, {}],
])
def test_augmentations_wont_change_float_input(augmentation_cls, params, float_image):
    float_image_copy = float_image.copy()
    aug = augmentation_cls(p=1, **params)
    aug(image=float_image)
    assert np.array_equal(float_image, float_image_copy)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [Cutout, {}],
    [CoarseDropout, {}],
    [JpegCompression, {}],
    [RandomBrightnessContrast, {}],
    [Blur, {}],
    [MotionBlur, {}],
    [MedianBlur, {}],
    [GaussianBlur, {}],
    [GaussNoise, {}],
    [InvertImg, {}],
    [RandomGamma, {}],
    [VerticalFlip, {}],
    [HorizontalFlip, {}],
    [Flip, {}],
    [Transpose, {}],
    [RandomRotate90, {}],
    [Rotate, {}],
    [OpticalDistortion, {}],
    [GridDistortion, {}],
    [ElasticTransform, {}],
    [GaussNoise, {}],
    [ToFloat, {}],
    [FromFloat, {}],
    [RandomGridShuffle, {}],
    [Solarize, {}],
])
def test_augmentations_wont_change_shape_grayscale(augmentation_cls, params, image, mask):
    aug = augmentation_cls(p=1, **params)

    # Test for grayscale image
    image = np.zeros((224, 224), dtype=np.uint8)
    mask = np.zeros((224, 224))
    result = aug(image=image, mask=mask)
    assert np.array_equal(image.shape, result['image'].shape)
    assert np.array_equal(mask.shape, result['mask'].shape)

    # Test for grayscale image with dummy dim
    image_1ch = np.zeros((224, 224, 1), dtype=np.uint8)
    mask_1ch = np.zeros((224, 224, 1))

    result = aug(image=image_1ch, mask=mask_1ch)
    assert np.array_equal(image_1ch.shape, result['image'].shape)
    assert np.array_equal(mask_1ch.shape, result['mask'].shape)

    # Test for RGB image
    image_3ch = np.zeros((224, 224, 3), dtype=np.uint8)
    mask_3ch = np.zeros((224, 224, 3))

    result = aug(image=image_3ch, mask=mask_3ch)
    assert np.array_equal(image_3ch.shape, result['image'].shape)
    assert np.array_equal(mask_3ch.shape, result['mask'].shape)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [Cutout, {}],
    [CoarseDropout, {}],
    [JpegCompression, {}],
    [HueSaturationValue, {}],
    [RGBShift, {}],
    [RandomBrightnessContrast, {}],
    [Blur, {}],
    [MotionBlur, {}],
    [MedianBlur, {}],
    [GaussianBlur, {}],
    [GaussNoise, {}],
    [CLAHE, {}],
    [ChannelShuffle, {}],
    [InvertImg, {}],
    [RandomGamma, {}],
    [ToGray, {}],
    [VerticalFlip, {}],
    [HorizontalFlip, {}],
    [Flip, {}],
    [Transpose, {}],
    [RandomRotate90, {}],
    [Rotate, {}],
    [OpticalDistortion, {}],
    [GridDistortion, {}],
    [ElasticTransform, {}],
    [Normalize, {}],
    [GaussNoise, {}],
    [ToFloat, {}],
    [FromFloat, {}],
    [RandomSnow, {}],
    [RandomRain, {}],
    [RandomFog, {}],
    [RandomSunFlare, {}],
    [RandomShadow, {}],
    [ChannelDropout, {}],
    [ISONoise, {}],
    [RandomGridShuffle, {}],
    [Solarize, {}],
])
def test_augmentations_wont_change_shape_rgb(augmentation_cls, params, image, mask):
    aug = augmentation_cls(p=1, **params)

    # Test for RGB image
    image_3ch = np.zeros((224, 224, 3), dtype=np.uint8)
    mask_3ch = np.zeros((224, 224, 3))

    result = aug(image=image_3ch, mask=mask_3ch)
    assert np.array_equal(image_3ch.shape, result['image'].shape)
    assert np.array_equal(mask_3ch.shape, result['mask'].shape)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [RandomCropNearBBox, {'max_part_shift': 0.15}],
])
def test_image_only_crop_around_bbox_augmentation(augmentation_cls, params, image, mask):
    aug = augmentation_cls(p=1, **params)
    annotations = {'image': image, 'cropping_bbox': [-59, 77, 177, 231]}
    data = aug(**annotations)
    assert data['image'].dtype == np.uint8


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [PadIfNeeded, {'min_height': 514, 'min_width': 514,
                   'border_mode': cv2.BORDER_CONSTANT, 'value': 100, 'mask_value': 1}],
    [Rotate, {'border_mode': cv2.BORDER_CONSTANT, 'value': 100, 'mask_value': 1}],
    [ShiftScaleRotate, {'border_mode': cv2.BORDER_CONSTANT, 'value': 100, 'mask_value': 1}],
    [OpticalDistortion, {'border_mode': cv2.BORDER_CONSTANT, 'value': 100, 'mask_value': 1}],
    [ElasticTransform, {'border_mode': cv2.BORDER_CONSTANT, 'value': 100, 'mask_value': 1}],
    [GridDistortion, {'border_mode': cv2.BORDER_CONSTANT, 'value': 100, 'mask_value': 1}],
])
def test_mask_fill_value(augmentation_cls, params):
    random.seed(42)
    aug = augmentation_cls(p=1, **params)
    input = {'image': np.zeros((512, 512), dtype=np.uint8) + 100, 'mask': np.ones((512, 512))}
    output = aug(**input)
    assert (output['image'] == 100).all()
    assert (output['mask'] == 1).all()



import random

import cv2
import pytest
import numpy as np
import imgaug as ia

import albumentations as A
import albumentations.augmentations.functional as F


TEST_SEEDS = (0, 1, 42, 111, 9999)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.JpegCompression, {}],
    [A.HueSaturationValue, {}],
    [A.RGBShift, {}],
    [A.RandomBrightnessContrast, {}],
    [A.Blur, {}],
    [A.MotionBlur, {}],
    [A.MedianBlur, {}],
    [A.GaussianBlur, {}],
    [A.GaussNoise, {}],
    [A.CLAHE, {}],
    [A.ChannelShuffle, {}],
    [A.InvertImg, {}],
    [A.RandomGamma, {}],
    [A.ToGray, {}],
    [A.Cutout, {}],
    [A.CoarseDropout, {}],
    [A.RandomSnow, {}],
    [A.RandomRain, {}],
    [A.RandomFog, {}],
    [A.RandomSunFlare, {}],
    [A.RandomShadow, {}],
    [A.PadIfNeeded, {}],
    [A.VerticalFlip, {}],
    [A.HorizontalFlip, {}],
    [A.Flip, {}],
    [A.Transpose, {}],
    [A.RandomRotate90, {}],
    [A.Rotate, {}],
    [A.ShiftScaleRotate, {}],
    [A.OpticalDistortion, {}],
    [A.GridDistortion, {}],
    [A.ElasticTransform, {}],
    [A.ToFloat, {}],
    [A.Normalize, {}],
    [A.RandomBrightness, {}],
    [A.RandomContrast, {}],
    [A.RandomScale, {}],
    [A.SmallestMaxSize, {}],
    [A.LongestMaxSize, {}],
    [A.RandomGridShuffle, {}],
    [A.Solarize, {}],
])
@pytest.mark.parametrize('p', [0.5, 1])
@pytest.mark.parametrize('seed', TEST_SEEDS)
@pytest.mark.parametrize('always_apply', (False, True))
def test_augmentations_serialization(augmentation_cls, params, p, seed, image, mask, always_apply):
    aug = augmentation_cls(p=p, always_apply=always_apply, **params)
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    aug_data = aug(image=image, mask=mask)
    random.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, mask=mask)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['mask'], deserialized_aug_data['mask'])


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.JpegCompression, {'quality_lower': 10, 'quality_upper': 80}],
    [A.HueSaturationValue, {'hue_shift_limit': 70, 'sat_shift_limit': 95, 'val_shift_limit': 55}],
    [A.RGBShift, {'r_shift_limit': 70, 'g_shift_limit': 80, 'b_shift_limit': 40}],
    [A.RandomBrightnessContrast, {'brightness_limit': 0.5, 'contrast_limit': 0.8}],
    [A.Blur, {'blur_limit': 3}],
    [A.MotionBlur, {'blur_limit': 3}],
    [A.MedianBlur, {'blur_limit': 3}],
    [A.GaussianBlur, {'blur_limit': 3}],
    [A.GaussNoise, {'var_limit': (20, 90)}],
    [A.CLAHE, {'clip_limit': 2, 'tile_grid_size': (12, 12)}],
    [A.RandomGamma, {'gamma_limit': (10, 90)}],
    [A.Cutout, {'num_holes': 4, 'max_h_size': 4, 'max_w_size': 4}],
    [A.CoarseDropout, {'max_holes': 4, 'max_height': 4, 'max_width': 4}],
    [A.RandomSnow, {'snow_point_lower': 0.2, 'snow_point_upper': 0.4, 'brightness_coeff': 4}],
    [A.RandomRain, {
        'slant_lower': -5,
        'slant_upper': 5,
        'drop_length': 15,
        'drop_width': 2,
        'drop_color': (100, 100, 100),
        'blur_value': 3,
        'brightness_coefficient': 0.5,
        'rain_type': 'heavy'
    }],
    [A.RandomFog, {'fog_coef_lower': 0.2, 'fog_coef_upper': 0.8, 'alpha_coef': 0.11}],
    [A.RandomSunFlare, {
        'flare_roi': (0.1, 0.1, 0.9, 0.6),
        'angle_lower': 0.1,
        'angle_upper': 0.95,
        'num_flare_circles_lower': 7,
        'num_flare_circles_upper': 11,
        'src_radius': 300,
        'src_color': (200, 200, 200),
    }],
    [A.RandomShadow, {
        'shadow_roi': (0.1, 0.4, 0.9, 0.9),
        'num_shadows_lower': 2,
        'num_shadows_upper': 4,
        'shadow_dimension': 8,
    }],
    [A.PadIfNeeded, {'min_height': 512, 'min_width': 512, 'border_mode': cv2.BORDER_CONSTANT, 'value': (10, 10, 10)}],
    [A.Rotate, {
        'limit': 120,
        'interpolation': cv2.INTER_CUBIC,
        'border_mode': cv2.BORDER_CONSTANT,
        'value': (10, 10, 10),
    }],
    [A.ShiftScaleRotate, {
        'shift_limit': 0.2,
        'scale_limit': 0.2,
        'rotate_limit': 70,
        'interpolation': cv2.INTER_CUBIC,
        'border_mode': cv2.BORDER_CONSTANT,
        'value': (10, 10, 10),
    }],
    [A.OpticalDistortion, {
        'distort_limit': 0.2,
        'shift_limit': 0.2,
        'interpolation': cv2.INTER_CUBIC,
        'border_mode': cv2.BORDER_CONSTANT,
        'value': (10, 10, 10),
    }],
    [A.GridDistortion, {
        'num_steps': 10,
        'distort_limit': 0.5,
        'interpolation': cv2.INTER_CUBIC,
        'border_mode': cv2.BORDER_CONSTANT,
        'value': (10, 10, 10),
    }],
    [A.ElasticTransform, {
        'alpha': 2,
        'sigma': 25,
        'alpha_affine': 40,
        'interpolation': cv2.INTER_CUBIC,
        'border_mode': cv2.BORDER_CONSTANT,
        'value': (10, 10, 10),
    }],
    [A.CenterCrop, {'height': 10, 'width': 10}],
    [A.RandomCrop, {'height': 10, 'width': 10}],
    [A.RandomSizedCrop, {'min_max_height': (4, 8), 'height': 10, 'width': 10}],
    [A.Crop, {'x_max': 64, 'y_max': 64}],
    [A.ToFloat, {'max_value': 16536}],
    [A.Normalize, {'mean': (0.385, 0.356, 0.306), 'std': (0.129, 0.124, 0.125), 'max_pixel_value': 100.0}],
    [A.RandomBrightness, {'limit': 0.4}],
    [A.RandomContrast, {'limit': 0.4}],
    [A.RandomScale, {'scale_limit': 0.2, 'interpolation': cv2.INTER_CUBIC}],
    [A.Resize, {'height': 64, 'width': 64}],
    [A.SmallestMaxSize, {'max_size': 64, 'interpolation': cv2.INTER_CUBIC}],
    [A.LongestMaxSize, {'max_size': 128, 'interpolation': cv2.INTER_CUBIC}],
    [A.RandomGridShuffle, {'grid': (5, 5)}],
    [A.Solarize, {'threshold': 32}],
])
@pytest.mark.parametrize('p', [0.5, 1])
@pytest.mark.parametrize('seed', TEST_SEEDS)
@pytest.mark.parametrize('always_apply', (False, True))
def test_augmentations_serialization_with_custom_parameters(
    augmentation_cls,
    params,
    p,
    seed,
    image,
    mask,
    always_apply,
):
    aug = augmentation_cls(p=p, always_apply=always_apply, **params)
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    aug_data = aug(image=image, mask=mask)
    random.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, mask=mask)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['mask'], deserialized_aug_data['mask'])


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.JpegCompression, {}],
    [A.HueSaturationValue, {}],
    [A.RGBShift, {}],
    [A.RandomBrightnessContrast, {}],
    [A.Blur, {}],
    [A.MotionBlur, {}],
    [A.MedianBlur, {}],
    [A.GaussianBlur, {}],
    [A.GaussNoise, {}],
    [A.CLAHE, {}],
    [A.ChannelShuffle, {}],
    [A.InvertImg, {}],
    [A.RandomGamma, {}],
    [A.ToGray, {}],
    [A.Cutout, {}],
    [A.GaussNoise, {}],
    [A.RandomSnow, {}],
    [A.RandomRain, {}],
    [A.RandomFog, {}],
    [A.RandomSunFlare, {}],
    [A.RandomShadow, {}],
    [A.PadIfNeeded, {}],
    [A.VerticalFlip, {}],
    [A.HorizontalFlip, {}],
    [A.Flip, {}],
    [A.Transpose, {}],
    [A.RandomRotate90, {}],
    [A.Rotate, {}],
    [A.ShiftScaleRotate, {}],
    [A.CenterCrop, {'height': 10, 'width': 10}],
    [A.RandomCrop, {'height': 10, 'width': 10}],
    [A.RandomSizedCrop, {'min_max_height': (4, 8), 'height': 10, 'width': 10}],
    [A.Crop, {'x_max': 64, 'y_max': 64}],
    [A.FromFloat, {}],
    [A.ToFloat, {}],
    [A.Normalize, {}],
    [A.RandomBrightness, {}],
    [A.RandomContrast, {}],
    [A.RandomScale, {}],
    [A.Resize, {'height': 64, 'width': 64}],
    [A.SmallestMaxSize, {}],
    [A.LongestMaxSize, {}],
    [A.RandomSizedBBoxSafeCrop, {'height': 50, 'width': 50}],
    [A.Solarize, {}],
])
@pytest.mark.parametrize('p', [0.5, 1])
@pytest.mark.parametrize('seed', TEST_SEEDS)
@pytest.mark.parametrize('always_apply', (False, True))
def test_augmentations_for_bboxes_serialization(augmentation_cls, params, p, seed, image, bboxes, always_apply):
    aug = augmentation_cls(p=p, always_apply=always_apply, **params)
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    aug_data = aug(image=image, bboxes=bboxes)
    random.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, bboxes=bboxes)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['bboxes'], deserialized_aug_data['bboxes'])


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.JpegCompression, {}],
    [A.HueSaturationValue, {}],
    [A.RGBShift, {}],
    [A.RandomBrightnessContrast, {}],
    [A.Blur, {}],
    [A.MotionBlur, {}],
    [A.MedianBlur, {}],
    [A.GaussianBlur, {}],
    [A.GaussNoise, {}],
    [A.CLAHE, {}],
    [A.ChannelShuffle, {}],
    [A.InvertImg, {}],
    [A.RandomGamma, {}],
    [A.ToGray, {}],
    [A.Cutout, {}],
    [A.GaussNoise, {}],
    [A.RandomSnow, {}],
    [A.RandomRain, {}],
    [A.RandomFog, {}],
    [A.RandomSunFlare, {}],
    [A.RandomShadow, {}],
    [A.PadIfNeeded, {}],
    [A.VerticalFlip, {}],
    [A.HorizontalFlip, {}],
    [A.Flip, {}],
    [A.RandomRotate90, {}],
    [A.Rotate, {}],
    [A.ShiftScaleRotate, {}],
    [A.CenterCrop, {'height': 10, 'width': 10}],
    [A.RandomCrop, {'height': 10, 'width': 10}],
    [A.RandomSizedCrop, {'min_max_height': (4, 8), 'height': 10, 'width': 10}],
    [A.FromFloat, {}],
    [A.ToFloat, {}],
    [A.Normalize, {}],
    [A.RandomBrightness, {}],
    [A.RandomContrast, {}],
    [A.RandomScale, {}],
    [A.Solarize, {}],
])
@pytest.mark.parametrize('p', [0.5, 1])
@pytest.mark.parametrize('seed', TEST_SEEDS)
@pytest.mark.parametrize('always_apply', (False, True))
def test_augmentations_for_keypoints_serialization(augmentation_cls, params, p, seed, image, keypoints, always_apply):
    aug = augmentation_cls(p=p, always_apply=always_apply, **params)
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    aug_data = aug(image=image, keypoints=keypoints)
    random.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, keypoints=keypoints)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['keypoints'], deserialized_aug_data['keypoints'])


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.IAAEmboss, {}],
    [A.IAASuperpixels, {}],
    [A.IAASharpen, {}],
    [A.IAAAdditiveGaussianNoise, {}],
    [A.IAACropAndPad, {}],
    [A.IAAFliplr, {}],
    [A.IAAFlipud, {}],
    [A.IAAAffine, {}],
    [A.IAAPiecewiseAffine, {}],
    [A.IAAPerspective, {}],
])
@pytest.mark.parametrize('p', [0.5, 1])
@pytest.mark.parametrize('seed', TEST_SEEDS)
@pytest.mark.parametrize('always_apply', (False, True))
def test_imgaug_augmentations_serialization(augmentation_cls, params, p, seed, image, mask, always_apply):
    aug = augmentation_cls(p=p, always_apply=always_apply, **params)
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    ia.seed(seed)
    aug_data = aug(image=image, mask=mask)
    random.seed(seed)
    ia.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, mask=mask)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['mask'], deserialized_aug_data['mask'])


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.IAAEmboss, {}],
    [A.IAASuperpixels, {}],
    [A.IAASharpen, {}],
    [A.IAAAdditiveGaussianNoise, {}],
    [A.IAACropAndPad, {}],
    [A.IAAFliplr, {}],
    [A.IAAFlipud, {}],
    [A.IAAAffine, {}],
    [A.IAAPiecewiseAffine, {}],
    [A.IAAPerspective, {}],
])
@pytest.mark.parametrize('p', [0.5, 1])
@pytest.mark.parametrize('seed', TEST_SEEDS)
@pytest.mark.parametrize('always_apply', (False, True))
def test_imgaug_augmentations_for_bboxes_serialization(augmentation_cls, params, p, seed, image, bboxes, always_apply):
    aug = augmentation_cls(p=p, always_apply=always_apply, **params)
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    ia.seed(seed)
    aug_data = aug(image=image, bboxes=bboxes)
    random.seed(seed)
    ia.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, bboxes=bboxes)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['bboxes'], deserialized_aug_data['bboxes'])


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.IAAEmboss, {}],
    [A.IAASuperpixels, {}],
    [A.IAASharpen, {}],
    [A.IAAAdditiveGaussianNoise, {}],
    [A.IAACropAndPad, {}],
    [A.IAAFliplr, {}],
    [A.IAAFlipud, {}],
    [A.IAAAffine, {}],
    [A.IAAPiecewiseAffine, {}],
    [A.IAAPerspective, {}],
])
@pytest.mark.parametrize('p', [0.5, 1])
@pytest.mark.parametrize('seed', TEST_SEEDS)
@pytest.mark.parametrize('always_apply', (False, True))
def test_imgaug_augmentations_for_keypoints_serialization(
    augmentation_cls,
    params,
    p,
    seed,
    image,
    keypoints,
    always_apply,
):
    aug = augmentation_cls(p=p, always_apply=always_apply, **params)
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    ia.seed(seed)
    aug_data = aug(image=image, keypoints=keypoints)
    random.seed(seed)
    ia.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, keypoints=keypoints)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['keypoints'], deserialized_aug_data['keypoints'])


@pytest.mark.parametrize('p', [0.5, 1])
@pytest.mark.parametrize('seed', TEST_SEEDS)
@pytest.mark.parametrize('always_apply', (False, True))
def test_image_only_crop_around_bbox_augmentation_serialization(p, seed, image, always_apply):
    aug = A.RandomCropNearBBox(p=p, always_apply=always_apply, max_part_shift=0.15)
    annotations = {'image': image, 'cropping_bbox': [-59, 77, 177, 231]}
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    aug_data = aug(**annotations)
    random.seed(seed)
    deserialized_aug_data = deserialized_aug(**annotations)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])


def test_from_float_serialization(float_image):
    aug = A.FromFloat(p=1, dtype='uint8')
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    aug_data = aug(image=float_image)
    deserialized_aug_data = deserialized_aug(image=float_image)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])


@pytest.mark.parametrize('seed', TEST_SEEDS)
def test_transform_pipeline_serialization(seed, image, mask):
    aug = A.Compose([
        A.OneOrOther(
            A.Compose([
                A.Resize(1024, 1024),
                A.RandomSizedCrop(min_max_height=(256, 1024), height=512, width=512, p=1),
                A.OneOf([
                    A.RandomSizedCrop(min_max_height=(256, 512), height=384, width=384, p=0.5),
                    A.RandomSizedCrop(min_max_height=(256, 512), height=512, width=512, p=0.5),
                ])
            ]),
            A.Compose([
                A.Resize(1024, 1024),
                A.RandomSizedCrop(min_max_height=(256, 1025), height=256, width=256, p=1),
                A.OneOf([
                    A.HueSaturationValue(p=0.5),
                    A.RGBShift(p=0.7)
                ], p=1),
            ])
        ),
        A.HorizontalFlip(p=1),
        A.RandomBrightnessContrast(p=0.5)
    ])
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    aug_data = aug(image=image, mask=mask)
    random.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, mask=mask)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['mask'], deserialized_aug_data['mask'])


@pytest.mark.parametrize(['bboxes', 'bbox_format', 'labels'], [
    [[[20, 30, 40, 50]], 'coco', [1]],
    [[[20, 30, 40, 50, 99], [10, 40, 30, 20, 9]], 'coco', [1, 2]],
    [[[20, 30, 60, 80]], 'pascal_voc', [2]],
    [[[20, 30, 60, 80, 99]], 'pascal_voc', [1]],
])
@pytest.mark.parametrize('seed', TEST_SEEDS)
def test_transform_pipeline_serialization_with_bboxes(seed, image, bboxes, bbox_format, labels):
    aug = A.Compose([
        A.OneOrOther(
            A.Compose([
                A.RandomRotate90(),
                A.OneOf([
                    A.HorizontalFlip(p=0.5),
                    A.VerticalFlip(p=0.5),
                ])
            ]),
            A.Compose([
                A.Rotate(p=0.5),
                A.OneOf([
                    A.HueSaturationValue(p=0.5),
                    A.RGBShift(p=0.7)
                ], p=1),
            ])
        ),
        A.HorizontalFlip(p=1),
        A.RandomBrightnessContrast(p=0.5)
    ], bbox_params={'format': bbox_format, 'label_fields': ['labels']})
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    aug_data = aug(image=image, bboxes=bboxes, labels=labels)
    random.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, bboxes=bboxes, labels=labels)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['bboxes'], deserialized_aug_data['bboxes'])


@pytest.mark.parametrize(['keypoints', 'keypoint_format', 'labels'], [
    [[[20, 30, 40, 50]], 'xyas', [1]],
    [[[20, 30, 40, 50, 99], [10, 40, 30, 20, 9]], 'xy', [1, 2]],
    [[[20, 30, 60, 80]], 'yx', [2]],
    [[[20, 30, 60, 80, 99]], 'xys', [1]],
])
@pytest.mark.parametrize('seed', TEST_SEEDS)
def test_transform_pipeline_serialization_with_keypoints(seed, image, keypoints, keypoint_format, labels):
    aug = A.Compose([
        A.OneOrOther(
            A.Compose([
                A.RandomRotate90(),
                A.OneOf([
                    A.HorizontalFlip(p=0.5),
                    A.VerticalFlip(p=0.5),
                ])
            ]),
            A.Compose([
                A.Rotate(p=0.5),
                A.OneOf([
                    A.HueSaturationValue(p=0.5),
                    A.RGBShift(p=0.7)
                ], p=1),
            ])
        ),
        A.HorizontalFlip(p=1),
        A.RandomBrightnessContrast(p=0.5)
    ], keypoint_params={'format': keypoint_format, 'label_fields': ['labels']})
    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    aug_data = aug(image=image, keypoints=keypoints, labels=labels)
    random.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, keypoints=keypoints, labels=labels)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['keypoints'], deserialized_aug_data['keypoints'])


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.ChannelShuffle, {}],
    [A.GaussNoise, {}],
    [A.Cutout, {}],
    [A.JpegCompression, {}],
    [A.HueSaturationValue, {}],
    [A.RGBShift, {}],
    [A.RandomBrightnessContrast, {}],
    [A.Blur, {}],
    [A.MotionBlur, {}],
    [A.MedianBlur, {}],
    [A.CLAHE, {}],
    [A.InvertImg, {}],
    [A.RandomGamma, {}],
    [A.ToGray, {}],
    [A.VerticalFlip, {}],
    [A.HorizontalFlip, {}],
    [A.Flip, {}],
    [A.Transpose, {}],
    [A.RandomRotate90, {}],
    [A.Rotate, {}],
    [A.OpticalDistortion, {}],
    [A.GridDistortion, {}],
    [A.ElasticTransform, {}],
    [A.Normalize, {}],
    [A.ToFloat, {}],
    [A.FromFloat, {}],
    [A.RandomGridShuffle, {}],
    [A.Solarize, {}],
])
@pytest.mark.parametrize('seed', TEST_SEEDS)
def test_additional_targets_for_image_only_serialization(augmentation_cls, params, image, seed):
    aug = A.Compose(
        [augmentation_cls(always_apply=True, **params)],
        additional_targets={'image2': 'image'},
    )
    image2 = image.copy()

    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug)
    random.seed(seed)
    aug_data = aug(image=image, image2=image2)
    random.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, image2=image2)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['image2'], deserialized_aug_data['image2'])


@pytest.mark.parametrize('seed', TEST_SEEDS)
@pytest.mark.parametrize('p', [1])
def test_lambda_serialization(image, mask, bboxes, keypoints, seed, p):

    def vflip_image(image, **kwargs):
        return F.vflip(image)

    def vflip_mask(mask, **kwargs):
        return F.vflip(mask)

    def vflip_bbox(bbox, **kwargs):
        return F.bbox_vflip(bbox, **kwargs)

    def vflip_keypoint(keypoint, **kwargs):
        return F.keypoint_vflip(keypoint, **kwargs)

    aug = A.Lambda(name='vflip', image=vflip_image, mask=vflip_mask, bbox=vflip_bbox, keypoint=vflip_keypoint, p=p)

    serialized_aug = A.to_dict(aug)
    deserialized_aug = A.from_dict(serialized_aug, lambda_transforms={'vflip': aug})
    random.seed(seed)
    aug_data = aug(image=image, mask=mask, bboxes=bboxes, keypoints=keypoints)
    random.seed(seed)
    deserialized_aug_data = deserialized_aug(image=image, mask=mask, bboxes=bboxes, keypoints=keypoints)
    assert np.array_equal(aug_data['image'], deserialized_aug_data['image'])
    assert np.array_equal(aug_data['mask'], deserialized_aug_data['mask'])
    assert np.array_equal(aug_data['bboxes'], deserialized_aug_data['bboxes'])
    assert np.array_equal(aug_data['keypoints'], deserialized_aug_data['keypoints'])



from __future__ import division

import math
from functools import wraps
from warnings import warn

import cv2
import numpy as np
from scipy.ndimage.filters import gaussian_filter

from albumentations.augmentations.bbox_utils import denormalize_bbox, normalize_bbox

MAX_VALUES_BY_DTYPE = {
    np.dtype('uint8'): 255,
    np.dtype('uint16'): 65535,
    np.dtype('uint32'): 4294967295,
    np.dtype('float32'): 1.0,
}


def clip(img, dtype, maxval):
    return np.clip(img, 0, maxval).astype(dtype)


def clipped(func):
    @wraps(func)
    def wrapped_function(img, *args, **kwargs):
        dtype = img.dtype
        maxval = MAX_VALUES_BY_DTYPE.get(dtype, 1.0)
        return clip(func(img, *args, **kwargs), dtype, maxval)

    return wrapped_function


def preserve_shape(func):
    """
    Preserve shape of the image

    """
    @wraps(func)
    def wrapped_function(img, *args, **kwargs):
        shape = img.shape
        result = func(img, *args, **kwargs)
        result = result.reshape(shape)
        return result

    return wrapped_function


def preserve_channel_dim(func):
    """
    Preserve dummy channel dim.

    """
    @wraps(func)
    def wrapped_function(img, *args, **kwargs):
        shape = img.shape
        result = func(img, *args, **kwargs)
        if len(shape) == 3 and shape[-1] == 1 and len(result.shape) == 2:
            result = np.expand_dims(result, axis=-1)
        return result

    return wrapped_function


def is_rgb_image(image):
    return len(image.shape) == 3 and image.shape[-1] == 3


def is_grayscale_image(image):
    return (len(image.shape) == 2) or (len(image.shape) == 3 and image.shape[-1] == 1)


def is_multispectral_image(image):
    return len(image.shape) == 3 and image.shape[-1] not in [1, 3]


def non_rgb_warning(image):
    if not is_rgb_image(image):
        message = 'This transformation expects 3-channel images'
        if is_grayscale_image(image):
            message += '\nYou can convert your grayscale image to RGB using cv2.cvtColor(image, cv2.COLOR_GRAY2RGB))'
        if is_multispectral_image(image):  # Any image with a number of channels other than 1 and 3
            message += '\nThis transformation cannot be applied to multi-spectral images'

        raise ValueError(message)


def vflip(img):
    return np.ascontiguousarray(img[::-1, ...])


def hflip(img):
    return np.ascontiguousarray(img[:, ::-1, ...])


def hflip_cv2(img):
    return cv2.flip(img, 1)


@preserve_shape
def random_flip(img, code):
    return cv2.flip(img, code)


def transpose(img):
    return img.transpose(1, 0, 2) if len(img.shape) > 2 else img.transpose(1, 0)


def rot90(img, factor):
    img = np.rot90(img, factor)
    return np.ascontiguousarray(img)


def normalize(img, mean, std, max_pixel_value=255.0):
    mean = np.array(mean, dtype=np.float32)
    mean *= max_pixel_value

    std = np.array(std, dtype=np.float32)
    std *= max_pixel_value

    denominator = np.reciprocal(std, dtype=np.float32)

    img = img.astype(np.float32)
    img -= mean
    img *= denominator
    return img


def cutout(img, holes, fill_value=0):
    # Make a copy of the input image since we don't want to modify it directly
    img = img.copy()
    for x1, y1, x2, y2 in holes:
        img[y1: y2, x1: x2] = fill_value
    return img


@preserve_channel_dim
def rotate(img, angle, interpolation=cv2.INTER_LINEAR, border_mode=cv2.BORDER_REFLECT_101, value=None):
    height, width = img.shape[:2]
    matrix = cv2.getRotationMatrix2D((width / 2, height / 2), angle, 1.0)
    img = cv2.warpAffine(img, matrix, (width, height), flags=interpolation, borderMode=border_mode, borderValue=value)
    return img


@preserve_channel_dim
def scale(img, scale, interpolation=cv2.INTER_LINEAR):
    height, width = img.shape[:2]
    new_height, new_width = int(height * scale), int(width * scale)
    img = cv2.resize(img, (new_width, new_height), interpolation=interpolation)
    return img


@preserve_channel_dim
def resize(img, height, width, interpolation=cv2.INTER_LINEAR):
    img = cv2.resize(img, (width, height), interpolation=interpolation)
    return img


@preserve_channel_dim
def shift_scale_rotate(img, angle, scale, dx, dy, interpolation=cv2.INTER_LINEAR, border_mode=cv2.BORDER_REFLECT_101,
                       value=None):
    height, width = img.shape[:2]
    center = (width / 2, height / 2)
    matrix = cv2.getRotationMatrix2D(center, angle, scale)
    matrix[0, 2] += dx * width
    matrix[1, 2] += dy * height
    img = cv2.warpAffine(img, matrix, (width, height), flags=interpolation, borderMode=border_mode, borderValue=value)
    return img


def bbox_shift_scale_rotate(bbox, angle, scale, dx, dy, interpolation, rows, cols, **params):
    height, width = rows, cols
    center = (width / 2, height / 2)
    matrix = cv2.getRotationMatrix2D(center, angle, scale)
    matrix[0, 2] += dx * width
    matrix[1, 2] += dy * height
    x = np.array([bbox[0], bbox[2], bbox[2], bbox[0]])
    y = np.array([bbox[1], bbox[1], bbox[3], bbox[3]])
    ones = np.ones(shape=(len(x)))
    points_ones = np.vstack([x, y, ones]).transpose()
    points_ones[:, 0] *= width
    points_ones[:, 1] *= height
    tr_points = matrix.dot(points_ones.T).T
    tr_points[:, 0] /= width
    tr_points[:, 1] /= height
    return [min(tr_points[:, 0]), min(tr_points[:, 1]), max(tr_points[:, 0]), max(tr_points[:, 1])]


def keypoint_shift_scale_rotate(keypoint, angle, scale, dx, dy, rows, cols, **params):
    height, width = rows, cols
    center = (width / 2, height / 2)
    x, y, a, s = keypoint
    matrix = cv2.getRotationMatrix2D(center, angle, scale)
    matrix[0, 2] += dx * width
    matrix[1, 2] += dy * height
    x, y = cv2.transform(np.array([[[x, y]]]), matrix).squeeze()
    return [x, y, a + math.radians(angle), s * scale]


def crop(img, x_min, y_min, x_max, y_max):
    height, width = img.shape[:2]
    if x_max <= x_min or y_max <= y_min:
        raise ValueError(
            'We should have x_min < x_max and y_min < y_max. But we got'
            ' (x_min = {x_min}, y_min = {y_min}, x_max = {x_max}, y_max = {y_max})'.format(
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max
            )
        )

    if x_min < 0 or x_max > width or y_min < 0 or y_max > height:
        raise ValueError(
            'Values for crop should be non negative and equal or smaller than image sizes'
            '(x_min = {x_min}, y_min = {y_min}, x_max = {x_max}, y_max = {y_max}'
            'height = {height}, width = {width})'.format(
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                height=height,
                width=width
            )
        )

    return img[y_min:y_max, x_min:x_max]


def get_center_crop_coords(height, width, crop_height, crop_width):
    y1 = (height - crop_height) // 2
    y2 = y1 + crop_height
    x1 = (width - crop_width) // 2
    x2 = x1 + crop_width
    return x1, y1, x2, y2


def center_crop(img, crop_height, crop_width):
    height, width = img.shape[:2]
    if height < crop_height or width < crop_width:
        raise ValueError(
            'Requested crop size ({crop_height}, {crop_width}) is '
            'larger than the image size ({height}, {width})'.format(
                crop_height=crop_height,
                crop_width=crop_width,
                height=height,
                width=width,
            )
        )
    x1, y1, x2, y2 = get_center_crop_coords(height, width, crop_height, crop_width)
    img = img[y1:y2, x1:x2]
    return img


def get_random_crop_coords(height, width, crop_height, crop_width, h_start, w_start):
    y1 = int((height - crop_height) * h_start)
    y2 = y1 + crop_height
    x1 = int((width - crop_width) * w_start)
    x2 = x1 + crop_width
    return x1, y1, x2, y2


def random_crop(img, crop_height, crop_width, h_start, w_start):
    height, width = img.shape[:2]
    if height < crop_height or width < crop_width:
        raise ValueError(
            'Requested crop size ({crop_height}, {crop_width}) is '
            'larger than the image size ({height}, {width})'.format(
                crop_height=crop_height,
                crop_width=crop_width,
                height=height,
                width=width,
            )
        )
    x1, y1, x2, y2 = get_random_crop_coords(height, width, crop_height, crop_width, h_start, w_start)
    img = img[y1:y2, x1:x2]
    return img


def clamping_crop(img, x_min, y_min, x_max, y_max):
    h, w = img.shape[:2]
    if x_min < 0:
        x_min = 0
    if y_min < 0:
        y_min = 0
    if y_max >= h:
        y_max = h - 1
    if x_max >= w:
        x_max = w - 1
    return img[int(y_min):int(y_max), int(x_min):int(x_max)]


def shift_hsv(img, hue_shift, sat_shift, val_shift):
    dtype = img.dtype
    img = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)
    if dtype == np.uint8:
        img = img.astype(np.int32)
    hue, sat, val = cv2.split(img)
    hue = cv2.add(hue, hue_shift)
    hue = np.where(hue < 0, hue + 180, hue)
    hue = np.where(hue > 180, hue - 180, hue)
    hue = hue.astype(dtype)
    sat = clip(cv2.add(sat, sat_shift), dtype, 255 if dtype == np.uint8 else 1.0)
    val = clip(cv2.add(val, val_shift), dtype, 255 if dtype == np.uint8 else 1.0)
    img = cv2.merge((hue, sat, val)).astype(dtype)
    img = cv2.cvtColor(img, cv2.COLOR_HSV2RGB)
    return img


def solarize(img, threshold=128):
    """Invert all pixel values above a threshold.

    Args:
        img: The image to solarize.
        threshold: All pixels above this greyscale level are inverted.

    Returns:
        Solarized image.

    """
    dtype = img.dtype
    max_val = MAX_VALUES_BY_DTYPE[dtype]

    if dtype == np.dtype('uint8'):
        lut = [(i if i < threshold else max_val - i) for i in range(max_val + 1)]

        prev_shape = img.shape
        img = cv2.LUT(img, np.array(lut, dtype=dtype))

        if len(prev_shape) != len(img.shape):
            img = np.expand_dims(img, -1)
        return img

    result_img = img.copy()
    cond = img >= threshold
    result_img[cond] = max_val - result_img[cond]
    return result_img


@clipped
def shift_rgb(img, r_shift, g_shift, b_shift):
    if img.dtype == np.uint8:
        img = img.astype('int32')
        r_shift, g_shift, b_shift = np.int32(r_shift), np.int32(g_shift), np.int32(b_shift)
    else:
        # Make a copy of the input image since we don't want to modify it directly
        img = img.copy()
    img[..., 0] += r_shift
    img[..., 1] += g_shift
    img[..., 2] += b_shift
    return img


def clahe(img, clip_limit=2.0, tile_grid_size=(8, 8)):
    if img.dtype != np.uint8:
        raise TypeError('clahe supports only uint8 inputs')

    clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=tile_grid_size)

    if len(img.shape) == 2:
        img = clahe.apply(img)
    else:
        img = cv2.cvtColor(img, cv2.COLOR_RGB2LAB)
        img[:, :, 0] = clahe.apply(img[:, :, 0])
        img = cv2.cvtColor(img, cv2.COLOR_LAB2RGB)

    return img


@preserve_channel_dim
def pad(img, min_height, min_width, border_mode=cv2.BORDER_REFLECT_101, value=None):
    height, width = img.shape[:2]

    if height < min_height:
        h_pad_top = int((min_height - height) / 2.0)
        h_pad_bottom = min_height - height - h_pad_top
    else:
        h_pad_top = 0
        h_pad_bottom = 0

    if width < min_width:
        w_pad_left = int((min_width - width) / 2.0)
        w_pad_right = min_width - width - w_pad_left
    else:
        w_pad_left = 0
        w_pad_right = 0

    img = pad_with_params(img, h_pad_top, h_pad_bottom, w_pad_left, w_pad_right, border_mode, value)

    assert img.shape[0] == max(min_height, height)
    assert img.shape[1] == max(min_width, width)

    return img


@preserve_channel_dim
def pad_with_params(img, h_pad_top, h_pad_bottom, w_pad_left, w_pad_right, border_mode=cv2.BORDER_REFLECT_101,
                    value=None):
    img = cv2.copyMakeBorder(img, h_pad_top, h_pad_bottom, w_pad_left, w_pad_right, border_mode, value=value)
    return img


@preserve_shape
def blur(img, ksize):
    return cv2.blur(img, (ksize, ksize))


@preserve_shape
def gaussian_blur(img, ksize):
    # When sigma=0, it is computed as `sigma = 0.3*((ksize-1)*0.5 - 1) + 0.8`
    return cv2.GaussianBlur(img, (ksize, ksize), sigmaX=0)


def _func_max_size(img, max_size, interpolation, func):
    height, width = img.shape[:2]

    scale = max_size / float(func(width, height))

    if scale != 1.0:
        out_size = tuple(py3round(dim * scale) for dim in (width, height))
        img = cv2.resize(img, out_size, interpolation=interpolation)
    return img


@preserve_channel_dim
def longest_max_size(img, max_size, interpolation):
    return _func_max_size(img, max_size, interpolation, max)


@preserve_channel_dim
def smallest_max_size(img, max_size, interpolation):
    return _func_max_size(img, max_size, interpolation, min)


@preserve_shape
def median_blur(img, ksize):
    if img.dtype == np.float32 and ksize not in {3, 5}:
        raise ValueError(
            'Invalid ksize value {}. For a float32 image the only valid ksize values are 3 and 5'.format(ksize))
    return cv2.medianBlur(img, ksize)


@preserve_shape
def motion_blur(img, kernel):
    return cv2.filter2D(img, -1, kernel / np.sum(kernel))


@preserve_shape
def jpeg_compression(img, quality):
    input_dtype = img.dtype
    needs_float = False

    if input_dtype == np.float32:
        warn('Jpeg compression augmentation '
             'is most effective with uint8 inputs, '
             '{} is used as input.'.format(input_dtype),
             UserWarning)
        img = from_float(img, dtype=np.dtype('uint8'))
        needs_float = True
    elif input_dtype not in (np.uint8, np.float32):
        raise ValueError('Unexpected dtype {} for Jpeg augmentation'.format(input_dtype))

    _, encoded_img = cv2.imencode('.jpg', img, (cv2.IMWRITE_JPEG_QUALITY, quality))
    img = cv2.imdecode(encoded_img, cv2.IMREAD_UNCHANGED)

    if needs_float:
        img = to_float(img, max_value=255)
    return img


@preserve_shape
def add_snow(img, snow_point, brightness_coeff):
    """Bleaches out pixels, mitation snow.

    From https://github.com/UjjwalSaxena/Automold--Road-Augmentation-Library

    Args:
        img:
        snow_point:
        brightness_coeff:

    Returns:

    """
    non_rgb_warning(img)

    input_dtype = img.dtype
    needs_float = False

    snow_point *= 127.5  # = 255 / 2
    snow_point += 85  # = 255 / 3

    if input_dtype == np.float32:
        img = from_float(img, dtype=np.dtype('uint8'))
        needs_float = True
    elif input_dtype not in (np.uint8, np.float32):
        raise ValueError('Unexpected dtype {} for RandomSnow augmentation'.format(input_dtype))

    image_HLS = cv2.cvtColor(img, cv2.COLOR_RGB2HLS)
    image_HLS = np.array(image_HLS, dtype=np.float32)

    image_HLS[:, :, 1][image_HLS[:, :, 1] < snow_point] *= brightness_coeff

    image_HLS[:, :, 1] = clip(image_HLS[:, :, 1], np.uint8, 255)

    image_HLS = np.array(image_HLS, dtype=np.uint8)

    image_RGB = cv2.cvtColor(image_HLS, cv2.COLOR_HLS2RGB)

    if needs_float:
        image_RGB = to_float(image_RGB, max_value=255)

    return image_RGB


@preserve_shape
def add_rain(img, slant, drop_length, drop_width, drop_color, blur_value, brightness_coefficient, rain_drops):
    """

    From https://github.com/UjjwalSaxena/Automold--Road-Augmentation-Library

    Args:
        img (np.uint8):
        slant (int):
        drop_length:
        drop_width:
        drop_color:
        blur_value (int): rainy view are blurry
        brightness_coefficient (float): rainy days are usually shady
        rain_drops:

    Returns:

    """
    non_rgb_warning(img)

    input_dtype = img.dtype
    needs_float = False

    if input_dtype == np.float32:
        img = from_float(img, dtype=np.dtype('uint8'))
        needs_float = True
    elif input_dtype not in (np.uint8, np.float32):
        raise ValueError('Unexpected dtype {} for RandomSnow augmentation'.format(input_dtype))

    image = img.copy()

    for (rain_drop_x0, rain_drop_y0) in rain_drops:
        rain_drop_x1 = rain_drop_x0 + slant
        rain_drop_y1 = rain_drop_y0 + drop_length

        cv2.line(image, (rain_drop_x0, rain_drop_y0), (rain_drop_x1, rain_drop_y1), drop_color, drop_width)

    image = cv2.blur(image, (blur_value, blur_value))  # rainy view are blurry
    image_hls = cv2.cvtColor(image, cv2.COLOR_RGB2HLS).astype(np.float32)
    image_hls[:, :, 1] *= brightness_coefficient

    image_rgb = cv2.cvtColor(image_hls.astype(np.uint8), cv2.COLOR_HLS2RGB)

    if needs_float:
        image_rgb = to_float(image_rgb, max_value=255)

    return image_rgb


@preserve_shape
def add_fog(img, fog_coef, alpha_coef, haze_list):
    """Add fog to the image.

    From https://github.com/UjjwalSaxena/Automold--Road-Augmentation-Library

    Args:
        img (np.array):
        fog_coef (float):
        alpha_coef (float):
        haze_list (list):
    Returns:

    """
    non_rgb_warning(img)

    input_dtype = img.dtype
    needs_float = False

    if input_dtype == np.float32:
        img = from_float(img, dtype=np.dtype('uint8'))
        needs_float = True
    elif input_dtype not in (np.uint8, np.float32):
        raise ValueError('Unexpected dtype {} for RandomFog augmentation'.format(input_dtype))

    height, width = img.shape[:2]

    hw = max(int(width // 3 * fog_coef), 10)

    for haze_points in haze_list:
        x, y = haze_points
        overlay = img.copy()
        output = img.copy()
        alpha = alpha_coef * fog_coef
        rad = hw // 2
        point = (x + hw // 2, y + hw // 2)
        cv2.circle(overlay, point, int(rad), (255, 255, 255), -1)
        cv2.addWeighted(overlay, alpha, output, 1 - alpha, 0, output)

        img = output.copy()

    image_rgb = cv2.blur(img, (hw // 10, hw // 10))

    if needs_float:
        image_rgb = to_float(image_rgb, max_value=255)

    return image_rgb


@preserve_shape
def add_sun_flare(img, flare_center_x, flare_center_y, src_radius, src_color, circles):
    """Add sun flare.

    From https://github.com/UjjwalSaxena/Automold--Road-Augmentation-Library

    Args:
        img (np.array):
        flare_center_x (float):
        flare_center_y (float):
        src_radius:
        src_color (int, int, int):
        circles (list):

    Returns:

    """
    non_rgb_warning(img)

    input_dtype = img.dtype
    needs_float = False

    if input_dtype == np.float32:
        img = from_float(img, dtype=np.dtype('uint8'))
        needs_float = True
    elif input_dtype not in (np.uint8, np.float32):
        raise ValueError('Unexpected dtype {} for RandomSunFlareaugmentation'.format(input_dtype))

    overlay = img.copy()
    output = img.copy()

    for (alpha, (x, y), rad3, (r_color, g_color, b_color)) in circles:
        cv2.circle(overlay, (x, y), rad3, (r_color, g_color, b_color), -1)

        cv2.addWeighted(overlay, alpha, output, 1 - alpha, 0, output)

    point = (int(flare_center_x), int(flare_center_y))

    overlay = output.copy()
    num_times = src_radius // 10
    alpha = np.linspace(0.0, 1, num=num_times)
    rad = np.linspace(1, src_radius, num=num_times)
    for i in range(num_times):
        cv2.circle(overlay, point, int(rad[i]), src_color, -1)
        alp = alpha[num_times - i - 1] * alpha[num_times - i - 1] * alpha[num_times - i - 1]
        cv2.addWeighted(overlay, alp, output, 1 - alp, 0, output)

    image_rgb = output

    if needs_float:
        image_rgb = to_float(image_rgb, max_value=255)

    return image_rgb


@preserve_shape
def add_shadow(img, vertices_list):
    """Add shadows to the image.

    From https://github.com/UjjwalSaxena/Automold--Road-Augmentation-Library

    Args:
        img (np.array):
        vertices_list (list):

    Returns:

    """
    non_rgb_warning(img)
    input_dtype = img.dtype
    needs_float = False

    if input_dtype == np.float32:
        img = from_float(img, dtype=np.dtype('uint8'))
        needs_float = True
    elif input_dtype not in (np.uint8, np.float32):
        raise ValueError('Unexpected dtype {} for RandomSnow augmentation'.format(input_dtype))

    image_hls = cv2.cvtColor(img, cv2.COLOR_RGB2HLS)
    mask = np.zeros_like(img)

    # adding all shadow polygons on empty mask, single 255 denotes only red channel
    for vertices in vertices_list:
        cv2.fillPoly(mask, vertices, 255)

    # if red channel is hot, image's "Lightness" channel's brightness is lowered
    red_max_value_ind = mask[:, :, 0] == 255
    image_hls[:, :, 1][red_max_value_ind] = image_hls[:, :, 1][red_max_value_ind] * 0.5

    image_rgb = cv2.cvtColor(image_hls, cv2.COLOR_HLS2RGB)

    if needs_float:
        image_rgb = to_float(image_rgb, max_value=255)

    return image_rgb


@preserve_shape
def optical_distortion(img, k=0, dx=0, dy=0, interpolation=cv2.INTER_LINEAR, border_mode=cv2.BORDER_REFLECT_101,
                       value=None):
    """Barrel / pincushion distortion. Unconventional augment.

    Reference:
        |  https://stackoverflow.com/questions/6199636/formulas-for-barrel-pincushion-distortion
        |  https://stackoverflow.com/questions/10364201/image-transformation-in-opencv
        |  https://stackoverflow.com/questions/2477774/correcting-fisheye-distortion-programmatically
        |  http://www.coldvision.io/2017/03/02/advanced-lane-finding-using-opencv/
    """
    height, width = img.shape[:2]

    fx = width
    fy = width

    cx = width * 0.5 + dx
    cy = height * 0.5 + dy

    camera_matrix = np.array([[fx, 0, cx],
                              [0, fy, cy],
                              [0, 0, 1]], dtype=np.float32)

    distortion = np.array([k, k, 0, 0, 0], dtype=np.float32)
    map1, map2 = cv2.initUndistortRectifyMap(camera_matrix, distortion, None, None, (width, height), cv2.CV_32FC1)
    img = cv2.remap(img, map1, map2, interpolation=interpolation, borderMode=border_mode, borderValue=value)
    return img


@preserve_shape
def grid_distortion(img, num_steps=10, xsteps=[], ysteps=[], interpolation=cv2.INTER_LINEAR,
                    border_mode=cv2.BORDER_REFLECT_101, value=None):
    """
    Reference:
        http://pythology.blogspot.sg/2014/03/interpolation-on-regular-distorted-grid.html
    """
    height, width = img.shape[:2]

    x_step = width // num_steps
    xx = np.zeros(width, np.float32)
    prev = 0
    for idx, x in enumerate(range(0, width, x_step)):
        start = x
        end = x + x_step
        if end > width:
            end = width
            cur = width
        else:
            cur = prev + x_step * xsteps[idx]

        xx[start:end] = np.linspace(prev, cur, end - start)
        prev = cur

    y_step = height // num_steps
    yy = np.zeros(height, np.float32)
    prev = 0
    for idx, y in enumerate(range(0, height, y_step)):
        start = y
        end = y + y_step
        if end > height:
            end = height
            cur = height
        else:
            cur = prev + y_step * ysteps[idx]

        yy[start:end] = np.linspace(prev, cur, end - start)
        prev = cur

    map_x, map_y = np.meshgrid(xx, yy)
    map_x = map_x.astype(np.float32)
    map_y = map_y.astype(np.float32)
    img = cv2.remap(img, map_x, map_y, interpolation=interpolation, borderMode=border_mode, borderValue=value)
    return img


@preserve_shape
def elastic_transform(image, alpha, sigma, alpha_affine, interpolation=cv2.INTER_LINEAR,
                      border_mode=cv2.BORDER_REFLECT_101, value=None, random_state=None, approximate=False):
    """Elastic deformation of images as described in [Simard2003]_ (with modifications).
    Based on https://gist.github.com/erniejunior/601cdf56d2b424757de5

    .. [Simard2003] Simard, Steinkraus and Platt, "Best Practices for
         Convolutional Neural Networks applied to Visual Document Analysis", in
         Proc. of the International Conference on Document Analysis and
         Recognition, 2003.
    """
    if random_state is None:
        random_state = np.random.RandomState(1234)

    height, width = image.shape[:2]

    # Random affine
    center_square = np.float32((height, width)) // 2
    square_size = min((height, width)) // 3
    alpha = float(alpha)
    sigma = float(sigma)
    alpha_affine = float(alpha_affine)

    pts1 = np.float32([center_square + square_size, [center_square[0] + square_size, center_square[1] - square_size],
                       center_square - square_size])
    pts2 = pts1 + random_state.uniform(-alpha_affine, alpha_affine, size=pts1.shape).astype(np.float32)
    matrix = cv2.getAffineTransform(pts1, pts2)

    image = cv2.warpAffine(image, matrix, (width, height), flags=interpolation, borderMode=border_mode,
                           borderValue=value)

    if approximate:
        # Approximate computation smooth displacement map with a large enough kernel.
        # On large images (512+) this is approximately 2X times faster
        dx = (random_state.rand(height, width).astype(np.float32) * 2 - 1)
        cv2.GaussianBlur(dx, (17, 17), sigma, dst=dx)
        dx *= alpha

        dy = (random_state.rand(height, width).astype(np.float32) * 2 - 1)
        cv2.GaussianBlur(dy, (17, 17), sigma, dst=dy)
        dy *= alpha
    else:
        dx = np.float32(gaussian_filter((random_state.rand(height, width) * 2 - 1), sigma) * alpha)
        dy = np.float32(gaussian_filter((random_state.rand(height, width) * 2 - 1), sigma) * alpha)

    x, y = np.meshgrid(np.arange(width), np.arange(height))

    mapx = np.float32(x + dx)
    mapy = np.float32(y + dy)

    return cv2.remap(image, mapx, mapy, interpolation, borderMode=border_mode, borderValue=value)


@preserve_shape
def elastic_transform_approx(image, alpha, sigma, alpha_affine, interpolation=cv2.INTER_LINEAR,
                             border_mode=cv2.BORDER_REFLECT_101, value=None, random_state=None):
    """Elastic deformation of images as described in [Simard2003]_ (with modifications for speed).
    Based on https://gist.github.com/erniejunior/601cdf56d2b424757de5

    .. [Simard2003] Simard, Steinkraus and Platt, "Best Practices for
         Convolutional Neural Networks applied to Visual Document Analysis", in
         Proc. of the International Conference on Document Analysis and
         Recognition, 2003.
    """
    if random_state is None:
        random_state = np.random.RandomState(1234)

    height, width = image.shape[:2]

    # Random affine
    center_square = np.float32((height, width)) // 2
    square_size = min((height, width)) // 3
    alpha = float(alpha)
    sigma = float(sigma)
    alpha_affine = float(alpha_affine)

    pts1 = np.float32([center_square + square_size, [center_square[0] + square_size, center_square[1] - square_size],
                       center_square - square_size])
    pts2 = pts1 + random_state.uniform(-alpha_affine, alpha_affine, size=pts1.shape).astype(np.float32)
    matrix = cv2.getAffineTransform(pts1, pts2)

    image = cv2.warpAffine(image, matrix, (width, height), flags=interpolation,
                           borderMode=border_mode, value=value)

    dx = (random_state.rand(height, width).astype(np.float32) * 2 - 1)
    cv2.GaussianBlur(dx, (17, 17), sigma, dst=dx)
    dx *= alpha

    dy = (random_state.rand(height, width).astype(np.float32) * 2 - 1)
    cv2.GaussianBlur(dy, (17, 17), sigma, dst=dy)
    dy *= alpha

    x, y = np.meshgrid(np.arange(width), np.arange(height))

    mapx = np.float32(x + dx)
    mapy = np.float32(y + dy)

    return cv2.remap(image, mapx, mapy, interpolation, borderMode=border_mode, borderValue=value)


def invert(img):
    return 255 - img


def channel_shuffle(img, channels_shuffled):
    img = img[..., channels_shuffled]
    return img


@preserve_shape
def channel_dropout(img, channels_to_drop, fill_value=0):
    if len(img.shape) == 2 or img.shape[2] == 1:
        raise NotImplementedError("Only one channel. ChannelDropout is not defined.")

    img = img.copy()

    img[..., channels_to_drop] = fill_value

    return img


@preserve_shape
def gamma_transform(img, gamma):
    if img.dtype == np.uint8:
        invGamma = 1.0 / gamma
        table = np.array([((i / 255.0) ** invGamma) * 255 for i in np.arange(0, 256)]).astype("uint8")
        img = cv2.LUT(img, table)
    else:
        img = np.power(img, gamma)

    return img


@clipped
def gauss_noise(image, gauss):
    image = image.astype('float32')
    return image + gauss


def _brightness_contrast_adjust_non_uint(img, alpha=1, beta=0):
    img = img.astype('float32')

    if alpha != 1:
        img *= alpha
    if beta != 0:
        img += beta * np.mean(img)
    return img


@preserve_shape
def _brightness_contrast_adjust_uint(img, alpha=1, beta=0):
    dtype = np.dtype('uint8')

    max_value = MAX_VALUES_BY_DTYPE[dtype]

    lut = np.arange(0, max_value + 1).astype('float32')

    if alpha != 1:
        lut *= alpha
    if beta != 0:
        lut += beta * np.mean(img)

    lut = np.clip(lut, 0, max_value).astype(dtype)
    img = cv2.LUT(img, lut)
    return img


@clipped
def brightness_contrast_adjust(img, alpha=1, beta=0):
    if img.dtype == np.uint8:
        return _brightness_contrast_adjust_uint(img, alpha, beta)
    else:
        return _brightness_contrast_adjust_non_uint(img, alpha, beta)


@clipped
def iso_noise(image, color_shift=0.05, intensity=0.5, random_state=None, **kwargs):
    """
    Apply poisson noise to image to simulate camera sensor noise.

    Args:
        image: Input image, currently, only RGB, uint8 images are supported.
        intensity: Multiplication factor for noise values. Values of ~0.5 are produce noticeable,
                   yet acceptable level of noise.
        random_state:
        **kwargs:

    Returns:
        Noised image

    """
    assert image.dtype == np.uint8, 'Image must have uint8 channel type'
    assert image.shape[2] == 3, 'Image must be RGB'

    if random_state is None:
        random_state = np.random.RandomState(42)

    one_over_255 = float(1. / 255.)
    image = np.multiply(image, one_over_255, dtype=np.float32)
    hls = cv2.cvtColor(image, cv2.COLOR_RGB2HLS)
    mean, stddev = cv2.meanStdDev(hls)

    luminance_noise = random_state.poisson(stddev[1] * intensity * 255, size=hls.shape[:2])
    color_noise = random_state.normal(0, color_shift * 360 * intensity, size=hls.shape[:2])

    hue = hls[..., 0]
    hue += color_noise
    hue[hue < 0] += 360
    hue[hue > 360] -= 360

    luminance = hls[..., 1]
    luminance += (luminance_noise / 255) * (1.0 - luminance)

    image = cv2.cvtColor(hls, cv2.COLOR_HLS2RGB) * 255
    return image.astype(np.uint8)


def to_gray(img):
    gray = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
    return cv2.cvtColor(gray, cv2.COLOR_GRAY2RGB)


def to_float(img, max_value=None):
    if max_value is None:
        try:
            max_value = MAX_VALUES_BY_DTYPE[img.dtype]
        except KeyError:
            raise RuntimeError(
                'Can\'t infer the maximum value for dtype {}. You need to specify the maximum value manually by '
                'passing the max_value argument'.format(img.dtype)
            )
    return img.astype('float32') / max_value


def from_float(img, dtype, max_value=None):
    if max_value is None:
        try:
            max_value = MAX_VALUES_BY_DTYPE[dtype]
        except KeyError:
            raise RuntimeError(
                'Can\'t infer the maximum value for dtype {}. You need to specify the maximum value manually by '
                'passing the max_value argument'.format(dtype)
            )
    return (img * max_value).astype(dtype)


def bbox_vflip(bbox, rows, cols):
    """Flip a bounding box vertically around the x-axis."""
    x_min, y_min, x_max, y_max = bbox
    return [x_min, 1 - y_max, x_max, 1 - y_min]


def bbox_hflip(bbox, rows, cols):
    """Flip a bounding box horizontally around the y-axis."""
    x_min, y_min, x_max, y_max = bbox
    return [1 - x_max, y_min, 1 - x_min, y_max]


def bbox_flip(bbox, d, rows, cols):
    """Flip a bounding box either vertically, horizontally or both depending on the value of `d`.

    Raises:
        ValueError: if value of `d` is not -1, 0 or 1.

    """
    if d == 0:
        bbox = bbox_vflip(bbox, rows, cols)
    elif d == 1:
        bbox = bbox_hflip(bbox, rows, cols)
    elif d == -1:
        bbox = bbox_hflip(bbox, rows, cols)
        bbox = bbox_vflip(bbox, rows, cols)
    else:
        raise ValueError('Invalid d value {}. Valid values are -1, 0 and 1'.format(d))
    return bbox


def crop_bbox_by_coords(bbox, crop_coords, crop_height, crop_width, rows, cols):
    """Crop a bounding box using the provided coordinates of bottom-left and top-right corners in pixels and the
    required height and width of the crop.
    """
    bbox = denormalize_bbox(bbox, rows, cols)
    x_min, y_min, x_max, y_max = bbox
    x1, y1, x2, y2 = crop_coords
    cropped_bbox = [x_min - x1, y_min - y1, x_max - x1, y_max - y1]
    return normalize_bbox(cropped_bbox, crop_height, crop_width)


def bbox_crop(bbox, x_min, y_min, x_max, y_max, rows, cols):
    crop_coords = [x_min, y_min, x_max, y_max]
    crop_height = y_max - y_min
    crop_width = x_max - x_min
    return crop_bbox_by_coords(bbox, crop_coords, crop_height, crop_width, rows, cols)


def bbox_center_crop(bbox, crop_height, crop_width, rows, cols):
    crop_coords = get_center_crop_coords(rows, cols, crop_height, crop_width)
    return crop_bbox_by_coords(bbox, crop_coords, crop_height, crop_width, rows, cols)


def bbox_random_crop(bbox, crop_height, crop_width, h_start, w_start, rows, cols):
    crop_coords = get_random_crop_coords(rows, cols, crop_height, crop_width, h_start, w_start)
    return crop_bbox_by_coords(bbox, crop_coords, crop_height, crop_width, rows, cols)


def bbox_rot90(bbox, factor, rows, cols):
    """Rotates a bounding box by 90 degrees CCW (see np.rot90)

    Args:
        bbox (tuple): A tuple (x_min, y_min, x_max, y_max).
        factor (int): Number of CCW rotations. Must be in range [0;3] See np.rot90.
        rows (int): Image rows.
        cols (int): Image cols.
    """
    if factor < 0 or factor > 3:
        raise ValueError('Parameter n must be in range [0;3]')
    x_min, y_min, x_max, y_max = bbox
    if factor == 1:
        bbox = [y_min, 1 - x_max, y_max, 1 - x_min]
    if factor == 2:
        bbox = [1 - x_max, 1 - y_max, 1 - x_min, 1 - y_min]
    if factor == 3:
        bbox = [1 - y_max, x_min, 1 - y_min, x_max]
    return bbox


def bbox_rotate(bbox, angle, rows, cols, interpolation):
    """Rotates a bounding box by angle degrees

    Args:
        bbox (tuple): A tuple (x_min, y_min, x_max, y_max).
        angle (int): Angle of rotation in degrees
        rows (int): Image rows.
        cols (int): Image cols.
        interpolation (int): interpolation method.

        return a tuple (x_min, y_min, x_max, y_max)
    """
    scale = cols / float(rows)
    x = np.array([bbox[0], bbox[2], bbox[2], bbox[0]])
    y = np.array([bbox[1], bbox[1], bbox[3], bbox[3]])
    x = x - 0.5
    y = y - 0.5
    angle = np.deg2rad(angle)
    x_t = (np.cos(angle) * x * scale + np.sin(angle) * y) / scale
    y_t = (-np.sin(angle) * x * scale + np.cos(angle) * y)
    x_t = x_t + 0.5
    y_t = y_t + 0.5
    return [min(x_t), min(y_t), max(x_t), max(y_t)]


def bbox_transpose(bbox, axis, rows, cols):
    """Transposes a bounding box along given axis.

    Args:
        bbox (tuple): A tuple (x_min, y_min, x_max, y_max).
        axis (int): 0 - main axis, 1 - secondary axis.
        rows (int): Image rows.
        cols (int): Image cols.
    """
    x_min, y_min, x_max, y_max = bbox
    if axis != 0 and axis != 1:
        raise ValueError('Axis must be either 0 or 1.')
    if axis == 0:
        bbox = [y_min, x_min, y_max, x_max]
    if axis == 1:
        bbox = [1 - y_max, 1 - x_max, 1 - y_min, 1 - x_min]
    return bbox


def keypoint_vflip(kp, rows, cols):
    """Flip a keypoint vertically around the x-axis."""
    x, y, angle, scale = kp
    c = math.cos(angle)
    s = math.sin(angle)
    angle = math.atan2(-s, c)
    return [x, (rows - 1) - y, angle, scale]


def keypoint_hflip(kp, rows, cols):
    """Flip a keypoint horizontally around the y-axis."""
    x, y, angle, scale = kp
    c = math.cos(angle)
    s = math.sin(angle)
    angle = math.atan2(s, -c)
    return [(cols - 1) - x, y, angle, scale]


def keypoint_flip(bbox, d, rows, cols):
    """Flip a keypoint either vertically, horizontally or both depending on the value of `d`.

    Raises:
        ValueError: if value of `d` is not -1, 0 or 1.

    """
    if d == 0:
        bbox = keypoint_vflip(bbox, rows, cols)
    elif d == 1:
        bbox = keypoint_hflip(bbox, rows, cols)
    elif d == -1:
        bbox = keypoint_hflip(bbox, rows, cols)
        bbox = keypoint_vflip(bbox, rows, cols)
    else:
        raise ValueError('Invalid d value {}. Valid values are -1, 0 and 1'.format(d))
    return bbox


def keypoint_rot90(keypoint, factor, rows, cols, **params):
    """Rotates a keypoint by 90 degrees CCW (see np.rot90)

    Args:
        keypoint (tuple): A tuple (x, y, angle, scale).
        factor (int): Number of CCW rotations. Must be in range [0;3] See np.rot90.
        rows (int): Image rows.
        cols (int): Image cols.
    """
    if factor < 0 or factor > 3:
        raise ValueError('Parameter n must be in range [0;3]')
    x, y, angle, scale = keypoint
    if factor == 1:
        keypoint = [y, (cols - 1) - x, angle - math.pi / 2, scale]
    if factor == 2:
        keypoint = [(cols - 1) - x, (rows - 1) - y, angle - math.pi, scale]
    if factor == 3:
        keypoint = [(rows - 1) - y, x, angle + math.pi / 2, scale]
    return keypoint


def keypoint_rotate(keypoint, angle, rows, cols, **params):
    matrix = cv2.getRotationMatrix2D(((cols - 1) * 0.5, (rows - 1) * 0.5), angle, 1.0)
    x, y, a, s = keypoint
    x, y = cv2.transform(np.array([[[x, y]]]), matrix).squeeze()
    return [x, y, a + math.radians(angle), s]


def keypoint_scale(keypoint, scale_x, scale_y, **params):
    """Scales a keypoint by scale_x and scale_y."""
    x, y, a, s = keypoint
    return [x * scale_x, y * scale_y, a, s * max(scale_x, scale_y)]


def crop_keypoint_by_coords(keypoint, crop_coords, crop_height, crop_width, rows, cols):
    """Crop a keypoint using the provided coordinates of bottom-left and top-right corners in pixels and the
    required height and width of the crop.
    """
    x, y, a, s = keypoint
    x1, y1, x2, y2 = crop_coords
    cropped_keypoint = [x - x1, y - y1, a, s]
    return cropped_keypoint


def keypoint_random_crop(keypoint, crop_height, crop_width, h_start, w_start, rows, cols):
    crop_coords = get_random_crop_coords(rows, cols, crop_height, crop_width, h_start, w_start)
    return crop_keypoint_by_coords(keypoint, crop_coords, crop_height, crop_width, rows, cols)


def keypoint_center_crop(bbox, crop_height, crop_width, rows, cols):
    crop_coords = get_center_crop_coords(rows, cols, crop_height, crop_width)
    return crop_keypoint_by_coords(bbox, crop_coords, crop_height, crop_width, rows, cols)


def py3round(number):
    """Unified rounding in all python versions."""
    if abs(round(number) - number) == 0.5:
        return int(2.0 * round(number / 2.0))

    return int(round(number))


def noop(input_obj, **params):
    return input_obj


def swap_tiles_on_image(image, tiles):
    """
    Swap tiles on image.

    Args:
        image (np.ndarray): Input image.
        tiles (np.ndarray): array of tuples(current_left_up_corner_row, current_left_up_corner_col,
                                            old_left_up_corner_row, old_left_up_corner_col,
                                            height_tile, width_tile)
    """
    new_image = image.copy()

    for idx, tile in enumerate(tiles):
        new_image[tile[0]:tile[0] + tile[4], tile[1]:tile[1] + tile[5]] = \
            image[tile[2]:tile[2] + tile[4], tile[3]:tile[3] + tile[5]]

    return new_image



import math

import numpy as np
import pytest
from albumentations import Flip, HorizontalFlip, VerticalFlip, IAAFliplr, IAAFlipud, CenterCrop

from albumentations.augmentations.keypoints_utils import convert_keypoint_from_albumentations, \
    convert_keypoints_from_albumentations, convert_keypoint_to_albumentations, convert_keypoints_to_albumentations
from albumentations.core.composition import Compose
from albumentations.core.transforms_interface import NoOp
from albumentations.augmentations.transforms import RandomSizedCrop
import albumentations.augmentations.functional as F


@pytest.mark.parametrize(['kp', 'source_format', 'expected'], [
    [[20, 30], 'xy', [20, 30, 0, 0]],
    [[20, 30], 'yx', [30, 20, 0, 0]],
    [[20, 30, 60], 'xys', [20, 30, 0, 60]],
    [[20, 30, 60], 'xya', [20, 30, math.radians(60), 0]],
    [[20, 30, 60, 80], 'xyas', [20, 30, math.radians(60), 80]],
])
def test_convert_keypoint_to_albumentations(kp, source_format, expected):
    image = np.ones((100, 100, 3))

    converted_keypoint = convert_keypoint_to_albumentations(kp, rows=image.shape[0], cols=image.shape[1],
                                                            source_format=source_format)
    assert converted_keypoint == expected


@pytest.mark.parametrize(['kp', 'target_format', 'expected'], [
    [[20, 30, 0, 0], 'xy', [20, 30]],
    [[20, 30, 0, 0], 'yx', [30, 20]],
    [[20, 30, 0.6, 0], 'xya', [20, 30, math.degrees(0.6)]],
    [[20, 30, 0, 0.6], 'xys', [20, 30, 0.6]],
    [[20, 30, 0.6, 80], 'xyas', [20, 30, math.degrees(0.6), 80]],
])
def test_convert_keypoint_from_albumentations(kp, target_format, expected):
    image = np.ones((100, 100, 3))
    converted_keypoint = convert_keypoint_from_albumentations(kp, rows=image.shape[0], cols=image.shape[1],
                                                              target_format=target_format)
    assert converted_keypoint == expected


@pytest.mark.parametrize(['kp', 'keypoint_format'], [
    [[20, 30, 40, 50], 'xy'],
    [[20, 30, 40, 50, 99], 'xyas'],
    [[20, 30, 60, 80], 'xysa'],
    [[20, 30, 60, 80, 99], 'yx'],
])
def test_convert_keypoint_to_albumentations_and_back(kp, keypoint_format):
    image = np.ones((100, 100, 3))
    converted_kp = convert_keypoint_to_albumentations(kp, rows=image.shape[0], cols=image.shape[1],
                                                      source_format=keypoint_format)
    converted_back_kp = convert_keypoint_from_albumentations(converted_kp, rows=image.shape[0], cols=image.shape[1],
                                                             target_format=keypoint_format)
    assert converted_back_kp == kp


def test_convert_keypoints_to_albumentations():
    keypoints = [[20, 30, 40, 50],
                 [30, 40, 50, 60, 99]]
    image = np.ones((100, 100, 3))
    converted_keypoints = convert_keypoints_to_albumentations(keypoints, rows=image.shape[0], cols=image.shape[1],
                                                              source_format='xyas')
    converted_keypoint_1 = convert_keypoint_to_albumentations(keypoints[0], rows=image.shape[0], cols=image.shape[1],
                                                              source_format='xyas')
    converted_keypoint_2 = convert_keypoint_to_albumentations(keypoints[1], rows=image.shape[0], cols=image.shape[1],
                                                              source_format='xyas')
    assert converted_keypoints == [converted_keypoint_1, converted_keypoint_2]


def test_convert_keypoints_from_albumentations():
    keypoints = [[0.2, 0.3, 0.6, 0.8], [0.3, 0.4, 0.7, 0.9, 99]]
    image = np.ones((100, 100, 3))
    converted_keypointes = convert_keypoints_from_albumentations(keypoints, rows=image.shape[0], cols=image.shape[1],
                                                                 target_format='xyas')
    converted_keypoint_1 = convert_keypoint_from_albumentations(keypoints[0], rows=image.shape[0], cols=image.shape[1],
                                                                target_format='xyas')
    converted_keypoint_2 = convert_keypoint_from_albumentations(keypoints[1], rows=image.shape[0], cols=image.shape[1],
                                                                target_format='xyas')
    assert converted_keypointes == [converted_keypoint_1, converted_keypoint_2]


@pytest.mark.parametrize(['keypoints', 'keypoint_format', 'labels'], [
    [[[20, 30, 40, 50]], 'xyas', [1]],
    [[[20, 30, 40, 50, 99], [10, 40, 30, 20, 9]], 'xy', None],
    [[[20, 30, 60, 80]], 'yx', [2]],
    [[[20, 30, 60, 80, 99]], 'xys', None],
])
def test_compose_with_keypoint_noop(keypoints, keypoint_format, labels):
    image = np.ones((100, 100, 3))
    if labels is not None:
        aug = Compose([NoOp(p=1.)], keypoint_params={'format': keypoint_format, 'label_fields': ['labels']})
        transformed = aug(image=image, keypoints=keypoints, labels=labels)
    else:
        aug = Compose([NoOp(p=1.)], keypoint_params={'format': keypoint_format})
        transformed = aug(image=image, keypoints=keypoints)
    assert np.array_equal(transformed['image'], image)
    assert transformed['keypoints'] == keypoints


@pytest.mark.parametrize(['keypoints', 'keypoint_format'], [
    [[[20, 30, 40, 50]], 'xyas'],
])
def test_compose_with_keypoint_noop_error_label_fields(keypoints, keypoint_format):
    image = np.ones((100, 100, 3))
    aug = Compose([NoOp(p=1.)], keypoint_params={'format': keypoint_format, 'label_fields': 'class_id'})
    with pytest.raises(Exception):
        aug(image=image, keypoints=keypoints, cls_id=[0])


@pytest.mark.parametrize(['keypoints', 'keypoint_format', 'labels'], [
    [[[20, 30, 60, 80]], 'xy', {'label': [1]}],
    [[], 'xy', {}],
    [[], 'xy', {'label': []}],
    [[[20, 30, 60, 80]], 'xy', {'id': [3]}],
    [[[20, 30, 60, 80], [30, 40, 40, 50]], 'xy', {'id': [3, 1]}],
])
def test_compose_with_keypoint_noop_label_outside(keypoints, keypoint_format, labels):
    image = np.ones((100, 100, 3))
    aug = Compose([NoOp(p=1.)], keypoint_params={'format': keypoint_format, 'label_fields': list(labels.keys())})
    transformed = aug(image=image, keypoints=keypoints, **labels)
    assert np.array_equal(transformed['image'], image)
    assert transformed['keypoints'] == keypoints
    for k, v in labels.items():
        assert transformed[k] == v


def test_random_sized_crop_size():
    image = np.ones((100, 100, 3))
    keypoints = [[0.2, 0.3, 0.6, 0.8], [0.3, 0.4, 0.7, 0.9, 99]]
    aug = RandomSizedCrop((70, 90), 50, 50, p=1.)
    transformed = aug(image=image, keypoints=keypoints)
    assert transformed['image'].shape == (50, 50, 3)
    assert len(keypoints) == len(transformed['keypoints'])


@pytest.mark.parametrize(['aug', 'keypoints', 'expected'], [
    [HorizontalFlip, [[0, 0]], [[2, 0]]],
    [HorizontalFlip, [[2, 0]], [[0, 0]]],
    [HorizontalFlip, [[0, 2]], [[2, 2]]],
    [HorizontalFlip, [[2, 2]], [[0, 2]]],
    #
    [VerticalFlip, [[0, 0]], [[0, 2]]],
    [VerticalFlip, [[2, 0]], [[2, 2]]],
    [VerticalFlip, [[0, 2]], [[0, 0]]],
    [VerticalFlip, [[2, 2]], [[2, 0]]],
    #
    [HorizontalFlip, [[1, 1]], [[1, 1]]],
    [VerticalFlip, [[1, 1]], [[1, 1]]]
])
def test_keypoint_flips_transform_3x3(aug, keypoints, expected):
    transform = Compose([aug(p=1)], keypoint_params={'format': 'xy'})

    image = np.ones((3, 3, 3))
    transformed = transform(image=image, keypoints=keypoints, labels=np.ones(len(keypoints)))
    assert np.allclose(expected, transformed['keypoints'])


@pytest.mark.parametrize(['aug', 'keypoints', 'expected'], [
    [HorizontalFlip, [[20, 30, 0, 0]], [[79, 30, 180, 0]]],
    [HorizontalFlip, [[20, 30, 45, 0]], [[79, 30, 135, 0]]],
    [HorizontalFlip, [[20, 30, 90, 0]], [[79, 30, 90, 0]]],
    #
    [VerticalFlip, [[20, 30, 0, 0]], [[20, 69, 0, 0]]],
    [VerticalFlip, [[20, 30, 45, 0]], [[20, 69, 315, 0]]],
    [VerticalFlip, [[20, 30, 90, 0]], [[20, 69, 270, 0]]],
])
def test_keypoint_transform_format_xyas(aug, keypoints, expected):
    transform = Compose([aug(p=1)], keypoint_params={'format': 'xyas',
                                                     'angle_in_degrees': True, 'label_fields': ['labels']})

    image = np.ones((100, 100, 3))
    transformed = transform(image=image, keypoints=keypoints, labels=np.ones(len(keypoints)))
    assert np.allclose(expected, transformed['keypoints'])


@pytest.mark.parametrize(['aug', 'keypoints', 'expected'], [
    [IAAFliplr, [[20, 30, 0, 0]], [[79, 30, 0, 0]]],
    [IAAFliplr, [[20, 30, 45, 0]], [[79, 30, 45, 0]]],
    [IAAFliplr, [[20, 30, 90, 0]], [[79, 30, 90, 0]]],
    #
    [IAAFlipud, [[20, 30, 0, 0]], [[20, 69, 0, 0]]],
    [IAAFlipud, [[20, 30, 45, 0]], [[20, 69, 45, 0]]],
    [IAAFlipud, [[20, 30, 90, 0]], [[20, 69, 90, 0]]],
])
def test_keypoint_transform_format_xy(aug, keypoints, expected):
    transform = Compose([aug(p=1)], keypoint_params={'format': 'xy', 'label_fields': ['labels']})

    image = np.ones((100, 100, 3))
    transformed = transform(image=image, keypoints=keypoints, labels=np.ones(len(keypoints)))
    assert np.allclose(expected, transformed['keypoints'])


@pytest.mark.parametrize(['aug', 'keypoints', 'expected'], [
    [IAAFliplr, [[20, 30, 0, 0]], [[79, 30, 0, 0]]],
])
def test_iaa_transforms_emit_warning(aug, keypoints, expected):
    with pytest.warns(UserWarning, match='IAAFliplr transformation supports only \'xy\' keypoints augmentation'):
        Compose([aug(p=1)], keypoint_params={'format': 'xyas', 'label_fields': ['labels']})


@pytest.mark.parametrize(['keypoint', 'expected', 'factor'], [
    [[20, 30, math.pi / 2, 0], [20, 30, math.pi / 2, 0], 0],
    [[20, 30, math.pi / 2, 0], [30, 179, 0, 0], 1],
    [[20, 30, math.pi / 2, 0], [179, 69, -math.pi / 2, 0], 2],
    [[20, 30, math.pi / 2, 0], [69, 20, math.pi, 0], 3],
])
def test_keypoint_rotate90(keypoint, expected, factor):
    actual = F.keypoint_rot90(keypoint, factor, rows=100, cols=200)
    assert actual == expected


@pytest.mark.parametrize(['keypoint', 'expected', 'angle'], [
    [[20, 30, math.pi / 2, 0], [20, 30, math.pi / 2, 0], 0],
    [[20, 30, math.pi / 2, 0], [30, 79, math.pi, 0], 90],
    [[20, 30, math.pi / 2, 0], [79, 69, math.pi + math.pi / 2, 0], 180],
    [[20, 30, math.pi / 2, 0], [69, 20, math.pi * 2, 0], 270],
    [[0, 0, 0, 0], [99, 99, math.pi, 0], 180],
    [[99, 99, 0, 0], [0, 0, math.pi, 0], 180],
])
def test_keypoint_rotate(keypoint, expected, angle):
    actual = F.keypoint_rotate(keypoint, angle, rows=100, cols=100)
    np.testing.assert_allclose(actual, expected, atol=1e-7)


@pytest.mark.parametrize(['keypoint', 'expected', 'scale'], [
    [[0.0, 0.0, math.pi / 2, 1], [0.0, 0.0, math.pi / 2, 1], 1],
    [[0.0, 0.0, math.pi / 2, 1], [0.0, 0.0, math.pi / 2, 2], 2],
    [[0.0, 0.0, math.pi / 2, 1], [0.0, 0.0, math.pi / 2, 0.5], 0.5],
])
def test_keypoint_scale(keypoint, expected, scale):
    actual = F.keypoint_scale(keypoint, scale, scale)
    np.testing.assert_allclose(actual, expected, atol=1e-7)


@pytest.mark.parametrize(['keypoint', 'expected', 'angle', 'scale', 'dx', 'dy'], [
    [[50, 50, 0, 5], [120, 160, math.pi / 2, 10], 90, 2, 0.1, 0.1],
])
def test_keypoint_shift_scale_rotate(keypoint, expected, angle, scale, dx, dy):
    actual = F.keypoint_shift_scale_rotate(keypoint, angle, scale, dx, dy, rows=100, cols=200)
    np.testing.assert_allclose(actual, expected, rtol=1e-4)


def test_compose_with_additional_targets():
    image = np.ones((100, 100, 3))
    keypoints = [[10, 10], [50, 50]]
    kp1 = [[15, 15], [55, 55]]
    aug = Compose([CenterCrop(50, 50)], keypoint_params={'format': 'xy'}, additional_targets={'kp1': 'keypoints'})
    transformed = aug(image=image, keypoints=keypoints, kp1=kp1)
    assert transformed['keypoints'] == [[25, 25]]
    assert transformed['kp1'] == [[30, 30]]



from functools import partial
from multiprocessing.pool import Pool

import numpy as np
import cv2
import pytest

import albumentations as A
import albumentations.augmentations.functional as F
from .conftest import skip_appveyor


def test_transpose_both_image_and_mask():
    image = np.ones((8, 6, 3))
    mask = np.ones((8, 6))
    augmentation = A.Transpose(p=1)
    augmented = augmentation(image=image, mask=mask)
    assert augmented['image'].shape == (6, 8, 3)
    assert augmented['mask'].shape == (6, 8)


@pytest.mark.parametrize('interpolation', [cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_CUBIC])
def test_rotate_interpolation(interpolation):
    image = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
    mask = np.random.randint(low=0, high=2, size=(100, 100), dtype=np.uint8)
    aug = A.Rotate(limit=(45, 45), interpolation=interpolation, p=1)
    data = aug(image=image, mask=mask)
    expected_image = F.rotate(image, 45, interpolation=interpolation, border_mode=cv2.BORDER_REFLECT_101)
    expected_mask = F.rotate(mask, 45, interpolation=cv2.INTER_NEAREST, border_mode=cv2.BORDER_REFLECT_101)
    assert np.array_equal(data['image'], expected_image)
    assert np.array_equal(data['mask'], expected_mask)


@pytest.mark.parametrize('interpolation', [cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_CUBIC])
def test_shift_scale_rotate_interpolation(interpolation):
    image = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
    mask = np.random.randint(low=0, high=2, size=(100, 100), dtype=np.uint8)
    aug = A.ShiftScaleRotate(shift_limit=(0.2, 0.2), scale_limit=(1.1, 1.1), rotate_limit=(45, 45),
                             interpolation=interpolation, p=1)
    data = aug(image=image, mask=mask)
    expected_image = F.shift_scale_rotate(image, angle=45, scale=2.1, dx=0.2, dy=0.2, interpolation=interpolation,
                                          border_mode=cv2.BORDER_REFLECT_101)
    expected_mask = F.shift_scale_rotate(mask, angle=45, scale=2.1, dx=0.2, dy=0.2, interpolation=cv2.INTER_NEAREST,
                                         border_mode=cv2.BORDER_REFLECT_101)
    assert np.array_equal(data['image'], expected_image)
    assert np.array_equal(data['mask'], expected_mask)


@pytest.mark.parametrize('interpolation', [cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_CUBIC])
def test_optical_distortion_interpolation(interpolation):
    image = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
    mask = np.random.randint(low=0, high=2, size=(100, 100), dtype=np.uint8)
    aug = A.OpticalDistortion(distort_limit=(0.05, 0.05), shift_limit=(0, 0), interpolation=interpolation, p=1)
    data = aug(image=image, mask=mask)
    expected_image = F.optical_distortion(image, k=0.05, dx=0, dy=0, interpolation=interpolation,
                                          border_mode=cv2.BORDER_REFLECT_101)
    expected_mask = F.optical_distortion(mask, k=0.05, dx=0, dy=0, interpolation=cv2.INTER_NEAREST,
                                         border_mode=cv2.BORDER_REFLECT_101)
    assert np.array_equal(data['image'], expected_image)
    assert np.array_equal(data['mask'], expected_mask)


@pytest.mark.parametrize('interpolation', [cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_CUBIC])
def test_grid_distortion_interpolation(interpolation):
    image = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
    mask = np.random.randint(low=0, high=2, size=(100, 100), dtype=np.uint8)
    aug = A.GridDistortion(num_steps=1, distort_limit=(0.3, 0.3), interpolation=interpolation, p=1)
    data = aug(image=image, mask=mask)
    expected_image = F.grid_distortion(image, num_steps=1, xsteps=[1.3], ysteps=[1.3], interpolation=interpolation,
                                       border_mode=cv2.BORDER_REFLECT_101)
    expected_mask = F.grid_distortion(mask, num_steps=1, xsteps=[1.3], ysteps=[1.3], interpolation=cv2.INTER_NEAREST,
                                      border_mode=cv2.BORDER_REFLECT_101)
    assert np.array_equal(data['image'], expected_image)
    assert np.array_equal(data['mask'], expected_mask)


@pytest.mark.parametrize('interpolation', [cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_CUBIC])
def test_elastic_transform_interpolation(monkeypatch, interpolation):
    image = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
    mask = np.random.randint(low=0, high=2, size=(100, 100), dtype=np.uint8)
    monkeypatch.setattr('albumentations.augmentations.transforms.ElasticTransform.get_params',
                        lambda *_: {'random_state': 1111})
    aug = A.ElasticTransform(alpha=1, sigma=50, alpha_affine=50, interpolation=interpolation, p=1)
    data = aug(image=image, mask=mask)
    expected_image = F.elastic_transform(image, alpha=1, sigma=50, alpha_affine=50, interpolation=interpolation,
                                         border_mode=cv2.BORDER_REFLECT_101,
                                         random_state=np.random.RandomState(1111))
    expected_mask = F.elastic_transform(mask, alpha=1, sigma=50, alpha_affine=50,
                                        interpolation=cv2.INTER_NEAREST,
                                        border_mode=cv2.BORDER_REFLECT_101,
                                        random_state=np.random.RandomState(1111))
    assert np.array_equal(data['image'], expected_image)
    assert np.array_equal(data['mask'], expected_mask)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.ElasticTransform, {}],
    [A.GridDistortion, {}],
    [A.ShiftScaleRotate, {'rotate_limit': 45}],
    [A.RandomScale, {'scale_limit': 0.5}],
    [A.RandomSizedCrop, {'min_max_height': (80, 90), 'height': 100, 'width': 100}],
    [A.RandomSizedBBoxSafeCrop, {'height': 100, 'width': 100}],
    [A.LongestMaxSize, {'max_size': 50}],
    [A.Rotate, {}],
    [A.OpticalDistortion, {}],
    [A.IAAAffine, {'scale': 1.5}],
    [A.IAAPiecewiseAffine, {'scale': 1.5}],
    [A.IAAPerspective, {}],
])
def test_binary_mask_interpolation(augmentation_cls, params):
    """Checks whether transformations based on DualTransform does not introduce a mask interpolation artifacts"""
    aug = augmentation_cls(p=1, **params)
    image = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
    mask = np.random.randint(low=0, high=2, size=(100, 100), dtype=np.uint8)
    data = aug(image=image, mask=mask)
    assert np.array_equal(np.unique(data['mask']), np.array([0, 1]))


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.ElasticTransform, {}],
    [A.GridDistortion, {}],
    [A.ShiftScaleRotate, {'rotate_limit': 45}],
    [A.RandomScale, {'scale_limit': 0.5}],
    [A.RandomSizedCrop, {'min_max_height': (80, 90), 'height': 100, 'width': 100}],
    [A.RandomSizedBBoxSafeCrop, {'height': 100, 'width': 100}],
    [A.LongestMaxSize, {'max_size': 50}],
    [A.Rotate, {}],
    [A.Resize, {'height': 80, 'width': 90}],
    [A.Resize, {'height': 120, 'width': 130}],
    [A.OpticalDistortion, {}]
])
def test_semantic_mask_interpolation(augmentation_cls, params):
    """Checks whether transformations based on DualTransform does not introduce a mask interpolation artifacts.
    Note: IAAAffine, IAAPiecewiseAffine, IAAPerspective does not properly operate if mask has values other than {0;1}
    """
    aug = augmentation_cls(p=1, **params)
    image = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
    mask = np.random.randint(low=0, high=4, size=(100, 100), dtype=np.uint8) * 64

    data = aug(image=image, mask=mask)
    assert np.array_equal(np.unique(data['mask']), np.array([0, 64, 128, 192]))


def __test_multiprocessing_support_proc(args):
    x, transform = args
    return transform(image=x)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.ElasticTransform, {}],
    [A.GridDistortion, {}],
    [A.ShiftScaleRotate, {'rotate_limit': 45}],
    [A.RandomScale, {'scale_limit': 0.5}],
    [A.RandomSizedCrop, {'min_max_height': (80, 90), 'height': 100, 'width': 100}],
    [A.RandomSizedBBoxSafeCrop, {'height': 100, 'width': 100}],
    [A.LongestMaxSize, {'max_size': 50}],
    [A.Rotate, {}],
    [A.OpticalDistortion, {}],
    [A.IAAAffine, {'scale': 1.5}],
    [A.IAAPiecewiseAffine, {'scale': 1.5}],
    [A.IAAPerspective, {}],
    [A.IAASharpen, {}]
])
@skip_appveyor
def test_multiprocessing_support(augmentation_cls, params):
    """Checks whether we can use augmentations in multi-threaded environments"""
    aug = augmentation_cls(p=1, **params)
    image = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)

    pool = Pool(8)
    pool.map(__test_multiprocessing_support_proc, map(lambda x: (x, aug), [image] * 100))
    pool.close()
    pool.join()


def test_force_apply():
    """
    Unit test for https://github.com/albu/albumentations/issues/189
    """
    aug = A.Compose([
        A.OneOrOther(
            A.Compose([
                A.RandomSizedCrop(min_max_height=(256, 1025), height=512, width=512, p=1),
                A.OneOf([
                    A.RandomSizedCrop(min_max_height=(256, 512), height=384, width=384, p=0.5),
                    A.RandomSizedCrop(min_max_height=(256, 512), height=512, width=512, p=0.5),
                ])
            ]),
            A.Compose([
                A.RandomSizedCrop(min_max_height=(256, 1025), height=256, width=256, p=1),
                A.OneOf([
                    A.HueSaturationValue(p=0.5),
                    A.RGBShift(p=0.7)
                ], p=1),
            ])
        ),
        A.HorizontalFlip(p=1),
        A.RandomBrightnessContrast(p=0.5)
    ])

    res = aug(image=np.zeros((1248, 1248, 3), dtype=np.uint8))
    assert res['image'].shape[0] in (256, 384, 512)
    assert res['image'].shape[1] in (256, 384, 512)


@pytest.mark.parametrize(['augmentation_cls', 'params'], [
    [A.ChannelShuffle, {}],
    [A.GaussNoise, {}],
    [A.Cutout, {}],
    [A.CoarseDropout, {}],
    [A.JpegCompression, {}],
    [A.HueSaturationValue, {}],
    [A.RGBShift, {}],
    [A.RandomBrightnessContrast, {}],
    [A.Blur, {}],
    [A.MotionBlur, {}],
    [A.MedianBlur, {}],
    [A.CLAHE, {}],
    [A.InvertImg, {}],
    [A.RandomGamma, {}],
    [A.ToGray, {}],
    [A.VerticalFlip, {}],
    [A.HorizontalFlip, {}],
    [A.Flip, {}],
    [A.Transpose, {}],
    [A.RandomRotate90, {}],
    [A.Rotate, {}],
    [A.OpticalDistortion, {}],
    [A.GridDistortion, {}],
    [A.ElasticTransform, {}],
    [A.Normalize, {}],
    [A.ToFloat, {}],
    [A.FromFloat, {}],
    [A.ChannelDropout, {}],
    [A.Solarize, {}],
])
def test_additional_targets_for_image_only(augmentation_cls, params):
    aug = A.Compose(
        [augmentation_cls(always_apply=True, **params)], additional_targets={'image2': 'image'})
    for i in range(10):
        image1 = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
        image2 = image1.copy()
        res = aug(image=image1, image2=image2)
        aug1 = res['image']
        aug2 = res['image2']
        assert np.array_equal(aug1, aug2)


def test_lambda_transform():
    def negate_image(image, **kwargs):
        return -image

    def one_hot_mask(mask, num_channels, **kwargs):
        new_mask = np.eye(num_channels, dtype=np.uint8)[mask]
        return new_mask

    def vflip_bbox(bbox, **kwargs):
        return F.bbox_vflip(bbox, **kwargs)

    def vflip_keypoint(keypoint, **kwargs):
        return F.keypoint_vflip(keypoint, **kwargs)

    aug = A.Lambda(
        image=negate_image,
        mask=partial(one_hot_mask, num_channels=16),
        bbox=vflip_bbox,
        keypoint=vflip_keypoint,
        p=1,
    )

    output = aug(
        image=np.ones((10, 10, 3), dtype=np.float32),
        mask=np.tile(np.arange(0, 10), (10, 1)),
        bboxes=[[10, 15, 25, 35]],
        keypoints=[[20, 30, 40, 50]],
    )
    assert (output['image'] < 0).all()
    assert output['mask'].shape[2] == 16  # num_channels
    assert output['bboxes'] == [F.bbox_vflip([10, 15, 25, 35], 10, 10)]
    assert output['keypoints'] == [F.keypoint_vflip([20, 30, 40, 50], 10, 10)]


def test_channel_droput():
    img = np.ones((10, 10, 3), dtype=np.float32)

    aug = A.ChannelDropout(channel_drop_range=(1, 1), always_apply=True)  # Drop one channel

    transformed = aug(image=img)['image']

    assert sum([transformed[:, :, c].max() for c in range(img.shape[2])]) == 2

    aug = A.ChannelDropout(channel_drop_range=(2, 2), always_apply=True)  # Drop two channels
    transformed = aug(image=img)['image']

    assert sum([transformed[:, :, c].max() for c in range(img.shape[2])]) == 1



import numpy as np
import torch

import albumentations as A
from albumentations.pytorch.transforms import ToTensor


def test_torch_to_tensor_augmentations(image, mask):
    aug = ToTensor()
    data = aug(image=image, mask=mask, force_apply=True)
    assert data['image'].dtype == torch.float32
    assert data['mask'].dtype == torch.float32


def test_additional_targets_for_totensor():
    aug = A.Compose(
        [ToTensor(num_classes=4)], additional_targets={'image2': 'image', 'mask2': 'mask'})
    for i in range(10):
        image1 = np.random.randint(low=0, high=256, size=(100, 100, 3), dtype=np.uint8)
        image2 = image1.copy()
        mask1 = np.random.randint(low=0, high=256, size=(100, 100, 4), dtype=np.uint8)
        mask2 = mask1.copy()
        res = aug(image=image1, image2=image2, mask=mask1, mask2=mask2)
        assert np.array_equal(res['image'], res['image2'])
        assert np.array_equal(res['mask'], res['mask2'])
